##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Christopher M. Bruns
# Contributors: Peter Eastman, Robert McGibbon
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
#
# Portions of this code originate from the OpenMM molecular simulation toolkit,
# copyright (c) 2012 Stanford University, Christopher M. Bruns and Peter Eastman,
# and are distributed under the following terms:
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS, CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
# USE OR OTHER DEALINGS IN THE SOFTWARE.
##############################################################################


import sys
import warnings

import numpy as np

from mdtraj.core import element


def _overflow_residue_check(num_str, pdbstructure, curr_atom):
    """
    Function to check and guess what the current residue is because it's overflowed. Lifted
    from a previous commit (c724024).

    Parameters
    ----------
    num_str : str
        str to be converted to int. Represents the residue number.
    pdbstructure : PdbStructure
        The PdbStructure object associated with `num_str`.
    curr_atom : Atom
        The current Atom object the residue number is associated with.

    Returns
    -------
    int
        The residue number deciphered in base 10.
    """
    if (
        pdbstructure._current_model is None
        or pdbstructure._current_model._current_chain is None
        or pdbstructure._current_model._current_chain._current_residue is None
    ):
        # This is the first residue in the model.
        return pdbstructure._next_residue_number
    else:
        currentRes = pdbstructure._current_model._current_chain._current_residue
        if currentRes.name_with_spaces != curr_atom.residue_name_with_spaces:
            # The residue name has changed.
            return pdbstructure._next_residue_number
        elif curr_atom.name_with_spaces in currentRes.atoms_by_name:
            # There is already an atom with this name.
            return pdbstructure._next_residue_number
        else:
            return currentRes.number


_atom_num_initial_nondecimal_functions = {
    "186a0": (lambda s, y=None: int(s, base=16)),  # "hex"
    "A0000": (lambda s, y=None: (int(s[0], base=36) * 10**4 + int(s[1:], base=36))),  # "chimera"
    "*****": (lambda s, y: y._next_atom_number),  # "overflow", y is the PdbStructure
}

_residue_num_initial_nondecimal_functions = {
    "2710": (lambda s, y=None, z=None: int(s, base=16)),  # 'hex'
    "A000": (lambda s, y=None, z=None: (int(s[0], base=36) * 10**3 + int(s[1:], base=36))),  # 'chimera'
    "****": _overflow_residue_check,  # 'overflow'
}


def _check_overflow_eligibility(num_str, str_type="atom"):
    """
    Return True if it's an overflow type, else False.

    An overflow type is defined as any dictionary keys above or
    something that looks the chimera format ({A..Z}000). The
    latter check exists because if residue numbers or atom numbers skip
    around and doesn't contain the start key, it might actually
    not recognize the overflow. This only exists for chimera-type values
    because it has no chance of false negatives.

    Parameters
    ----------
    num_str : str
        str to be checked whether it's overflowed or not.
    str_type : str, default: 'atom'
        The type of num_str. 'atom' if it's for an atom number,
        'residue' if it's for a residue number.

    Returns
    -------
    bool
        True if overflow type, False if not.
    """
    # Check a different dictionary depending on the type.
    if str_type == "atom":
        if num_str in _atom_num_initial_nondecimal_functions:
            return True
    elif str_type == "residue":
        if num_str in _residue_num_initial_nondecimal_functions:
            return True

    # This is mostly to 'guess' if the number is a chimera-type overflow, in case we skipped
    if 65 <= ord(num_str[0]) <= 90:
        # ord({A..Z}) should be in [65..90]
        return True
    else:
        return False


def _read_atom_number(num_str, pdbstructure=None):
    """
    This function determines whether we need to swap to overflow mode. Otherwise, we'll just
    turn ``num_str`` into an integer.

    If it's in any of the non-decimal modes, we will attempt to set the _atom_num_nondec_mode to the
    correct key. With this set, all subsequent atom numbers will be deciphered using a corresponding
    function.

    Parameters
    ----------
    num_str : str
        str to be converted to int. Represents the atom number.
    pdbstructure : PdbStructure, default: None
        The PdbStructure object associated with `num_str`.

    Returns
    -------
    int
        The atom number deciphered in base 10.
    """
    try:
        if pdbstructure._atom_num_nondec_mode is not None:
            # If it already has an overflow function, then it will use the corresponding
            # _atom_num_function as dictated by pdbstructure._atom_num_nondec_mode to decipher the num_str.
            return pdbstructure._atom_num_nondec_mode(num_str, pdbstructure)
        elif pdbstructure._next_atom_number > 99999 and _check_overflow_eligibility(num_str, "atom"):
            # If the next atom number is > 99999 and our current atom number is one of the overflow keys,
            # raise an OverflowError, which will switch to the correct mode to read num_str.
            raise OverflowError("Need to parse atom number using non-decimal residue modes.")
        else:
            return int(num_str)
    except (AttributeError, ValueError, OverflowError, KeyError):
        # we need to figure out on the 1st try which mode to switch to. There are currently 3
        # options: VMD (hex), Chimera (their own 'hybrid36' mode), and overflow (*****).
        # Chimera starts with A0000, vmd with 186a0, so they are distinguishable.
        if pdbstructure is None:
            try:
                return int(num_str)
            except ValueError:
                # num_str is not decimal, no nondec_mode to interpret it, no pdbstructure to say
                # what it is or to provide current number of atoms. No way to figure out
                return 0
        else:
            if pdbstructure._atom_num_nondec_mode is None:
                try:
                    # If this is the first time we reached overflow, we will try to determine the mode
                    # using the num_str as key.
                    pdbstructure._atom_num_nondec_mode = _atom_num_initial_nondecimal_functions[num_str]
                except KeyError:
                    if 65 <= ord(num_str[0]) <= 90:
                        # Could be a chimera-type overflow? Attempting to guess with that assumption.
                        pdbstructure._atom_num_nondec_mode = _atom_num_initial_nondecimal_functions["A0000"]
            try:
                # The _atom_num_nondec_mode has been set.
                # Try and run the corresponding _atom_num_functions on num_str
                return pdbstructure._atom_num_nondec_mode(num_str, pdbstructure)
            except (ValueError, KeyError, TypeError):
                # Didn't work, we need to change to overflow mode and guess with _next_atom_number.
                pdbstructure._atom_num_nondec_mode = _atom_num_initial_nondecimal_functions["*****"]
                warnings.warn(
                    f"Need to guess atom number {num_str} starting from atom {pdbstructure._next_atom_number}.",
                )
                return pdbstructure._atom_num_nondec_mode(num_str, pdbstructure)


def _read_residue_number(num_str, pdbstructure=None, curr_atom=None):
    """
    This function determines whether we need to swap to overflow mode. Otherwise, we'll just
    turn ``num_str`` into an integer.

    If it's in any of the non-decimal modes, we will attempt to set the _residue_num_nondec_mode to the
    correct key. With this set, all subsequent residue numbers will be deciphered using a corresponding
    function. If it's in "overflow" mode, will try to guess the residue number.

    Parameters
    ----------
    num_str : str
        str to be converted to int. Represents the residue number.
    pdbstructure : PdbStructure, default: None
        The PdbStructure object associated with `num_str`.
    curr_atom : Atom, default: None
        The current Atom object the residue number is associated with.

    Returns
    -------
    int
        The residue number deciphered in base 10.
    """
    try:
        if pdbstructure._residue_num_nondec_mode is not None:
            # If it already has an overflow function, then it will use the corresponding _residue_num_function
            # as dictated by pdbstructure._residue_num_nondec_mode to decipher the num_str.
            return pdbstructure._residue_num_nondec_mode(num_str, pdbstructure, curr_atom)
        elif pdbstructure._next_residue_number > 9999 and _check_overflow_eligibility(num_str, "residue"):
            # If the next residue number is > 9999 and our current residue number is one of the nondecimal keys,
            # Raise an OverflowError, which will switch to the correct mode to read num_str.
            raise OverflowError("Need to parse residue number using non-decimal residue modes.")
        else:
            return int(num_str)
    except (AttributeError, OverflowError, KeyError, ValueError):
        # we need to figure out on the 1st try which mode to switch to. There are currently 3 options:
        # VMD (hex) and Chimera (their own 'hybrid36' mode) and Overflow (****).
        # Chimera starts with A000, vmd with 2710, and Overflow just shows ****.
        # The can be turned into decimal with "int()" so the "hex" mode will only be
        # activated when _next_residue_number > 9999 (maximum in decimal) and current num_str
        # isn't 9999.
        if pdbstructure is None:
            try:
                return int(num_str)
            except ValueError:
                # num_str is not decimal, no pdbstructure to say what it is or
                # to provide current number of atoms. No way to figure out
                return 0
        else:
            if pdbstructure._residue_num_nondec_mode is None:
                # Attempt to set overflow mode using num_str as key
                try:
                    pdbstructure._residue_num_nondec_mode = _residue_num_initial_nondecimal_functions[num_str]
                except KeyError:
                    if 65 <= ord(num_str[0]) <= 90:
                        # Could be a chimera-type overflow? Attempting to guess with that assumption.
                        pdbstructure._residue_num_nondec_mode = _residue_num_initial_nondecimal_functions["A000"]
            try:
                # Try and run the _residue_num_functions
                return pdbstructure._residue_num_nondec_mode(num_str, pdbstructure, curr_atom)
            except (ValueError, KeyError, TypeError):
                # Didn't work, we need to change to overflow mode and guess with _next_residue_number
                pdbstructure._residue_num_nondec_mode = _residue_num_initial_nondecimal_functions["****"]
                warnings.warn(
                    f"Need to guess residue number string {num_str} starting from "
                    f"residue{pdbstructure._next_residue_number}, atom "
                    f"{pdbstructure._next_atom_number}.",
                )
                return pdbstructure._residue_num_nondec_mode(num_str, pdbstructure, curr_atom)


class PdbStructure:
    """
    PdbStructure object holds a parsed Protein Data Bank format file.

    Examples:

    Load a pdb structure from a file:
    > pdb = PdbStructure(open("1ARJ.pdb"))

    Fetch the first atom of the structure:
    > print(pdb.iter_atoms().next())
    ATOM      1  O5'   G N  17      13.768  -8.431  11.865  1.00  0.00           O

    Loop over all of the atoms of the structure
    > for atom in pdb.iter_atoms():
    >     print(atom)
    ATOM      1  O5'   G N  17      13.768  -8.431  11.865  1.00  0.00           O
    ...

    Get a list of all atoms in the structure:
    > atoms = list(pdb.iter_atoms())

    also:
    residues = list(pdb.iter_residues())
    positions = list(pdb.iter_positions())
    chains = list(pdb.iter_chains())
    models = list(pdb.iter_models())

    Fetch atomic coordinates of first atom:
    > print(pdb.iter_positions().next())
    [13.768, -8.431, 11.865]

     or

    > print(pdb.iter_atoms().next().position)
    [13.768, -8.431, 11.865]

    Strip the length units from an atomic position:
    > pos = pdb.iter_positions().next()
    > print(pos)
    [13.768, -8.431, 11.865]
    > print(pos)
    [13.768, -8.431, 11.865]
    > print(pos)
    [1.3768, -0.8431, 1.1865]


    The hierarchical structure of the parsed PDB structure is as follows:
    PdbStructure
      Model
        Chain
          Residue
            Atom
              Location

    Model - A PDB structure consists of one or more Models.  Each model corresponds to one version of
    an NMR structure, or to one frame of a molecular dynamics trajectory.

    Chain - A Model contains one or more Chains.  Each chain corresponds to one molecule, although multiple
    water molecules are frequently included in the same chain.

    Residue - A Chain contains one or more Residues.  One Residue corresponds to one of the repeating
    unit that constitutes a polymer such as protein or DNA.  For non-polymeric molecules, one Residue
    represents one molecule.

    Atom - A Residue contains one or more Atoms.  Atoms are chemical atoms.

    Location - An atom can sometimes have more that one position, due to static disorder in X-ray
    crystal structures.  To see all of the atom positions, use the atom.iter_positions() method,
    or pass the parameter "include_alt_loc=True" to one of the other iter_positions() methods.

    > for pos in pdb.iter_positions(include_alt_loc=True):
    >   ...

    Will loop over all atom positions, including multiple alternate locations for atoms that have
    multiple positions.  The default value of include_alt_loc is False for the iter_positions()
    methods.
    """

    def __init__(self, input_stream, load_all_models=True):
        """Create a PDB model from a PDB file stream.

        Parameters:
         - self (PdbStructure) The new object that is created.
         - input_stream (stream) An input file stream, probably created with
             open().
         - load_all_models (bool) Whether to load every model of an NMR
             structure or trajectory, or just load the first model, to save memory.
        """
        self._atom_num_nondec_mode = None  # None (decimal until changes), 'hex', 'chimera', 'overflow'
        self._residue_num_nondec_mode = None  # None (decimal until changes), 'hex', 'chimera', 'overflow'

        # initialize models
        self.load_all_models = load_all_models
        self.models = []
        self._current_model = None
        self.default_model = None
        self.models_by_number = {}
        self._unit_cell_lengths = None
        self._unit_cell_angles = None
        # read file
        self._load(input_stream)

    def _load(self, input_stream):
        state = None

        self._reset_atom_numbers()
        self._reset_residue_numbers()

        # Read one line at a time
        for pdb_line in input_stream:
            # Look for atoms
            if (pdb_line.find("ATOM  ") == 0) or (pdb_line.find("HETATM") == 0):
                if state == "NEW_MODEL":
                    new_number = self._current_model.number + 1
                    self._add_model(Model(new_number))
                    state = None
                self._add_atom(Atom(pdb_line, self))
            # Notice MODEL punctuation, for the next level of detail
            # in the structure->model->chain->residue->atom->position hierarchy
            elif pdb_line.find("MODEL") == 0:
                # model_number = int(pdb_line[10:14])
                if self._current_model is None:
                    new_number = 0
                else:
                    new_number = self._current_model.number + 1
                self._add_model(Model(new_number))
                self._reset_atom_numbers()
                self._reset_residue_numbers()
                state = None

            elif pdb_line.find("ENDMDL") == 0:
                self._current_model._finalize()
                if self.load_all_models:
                    state = "NEW_MODEL"
                else:
                    break

            elif pdb_line.find("END") == 0:
                self._current_model._finalize()
                if self.load_all_models:
                    state = "NEW_MODEL"
                else:
                    break

            elif pdb_line.find("TER") == 0 and pdb_line.split()[0] == "TER":
                self._current_model._current_chain._add_ter_record()
                self._reset_residue_numbers()

            elif pdb_line.find("CRYST1") == 0:
                self._unit_cell_lengths = (
                    float(pdb_line[6:15]),
                    float(pdb_line[15:24]),
                    float(pdb_line[24:33]),
                )
                self._unit_cell_angles = (
                    float(pdb_line[33:40]),
                    float(pdb_line[40:47]),
                    float(pdb_line[47:54]),
                )

            elif pdb_line.find("CONECT") == 0:
                atoms = []
                # :-1 to remove '\n' in the end so rstrip can work, -5 to leave space for +5 in the 'pos : pos+5'
                ll = len(pdb_line[:-1].rstrip(" ")) - 5

                for pos in [p for p in [6, 11, 16, 21, 26] if (p <= ll)]:
                    atoms.append(_read_atom_number(pdb_line[pos : pos + 5], pdbstructure=self))

                self._current_model.connects.append(atoms)

        self._finalize()

    def _reset_atom_numbers(self):
        self._atom_num_nondec_mode = None  # None (decimal until changes), 'hex', 'chimera', 'overflow'
        self._next_atom_number = 1

    def _reset_residue_numbers(self):
        self._residue_num_nondec_mode = None  # None (decimal until changes), 'hex', 'chimera', 'overflow'
        self._next_residue_number = 1

    def write(self, output_stream=sys.stdout):
        """Write out structure in PDB format"""
        for model in self.models:
            if len(model.chains) == 0:
                continue
            if len(self.models) > 1:
                print("MODEL     %4d" % model.number, file=output_stream)
            model.write(output_stream)
            if len(self.models) > 1:
                print("ENDMDL", file=output_stream)
        print("END", file=output_stream)

    def _add_model(self, model):
        if self.default_model is None:
            self.default_model = model
        self.models.append(model)
        self._current_model = model
        if model.number not in self.models_by_number:
            self.models_by_number[model.number] = model

    def get_model(self, model_number):
        return self.models_by_number[model_number]

    def model_numbers(self):
        return list(self.models_by_number.keys())

    def __contains__(self, model_number):
        return self.models_by_number.__contains__(model_number)

    def __getitem__(self, model_number):
        return self.models_by_number[model_number]

    def __iter__(self):
        yield from self.models

    def iter_models(self, use_all_models=False):
        if use_all_models:
            yield from self
        elif len(self.models) > 0:
            yield self.models[0]

    def iter_chains(self, use_all_models=False):
        for model in self.iter_models(use_all_models):
            yield from model.iter_chains()

    def iter_residues(self, use_all_models=False):
        for model in self.iter_models(use_all_models):
            yield from model.iter_residues()

    def iter_atoms(self, use_all_models=False):
        for model in self.iter_models(use_all_models):
            yield from model.iter_atoms()

    def iter_positions(self, use_all_models=False, include_alt_loc=False):
        """
        Iterate over atomic positions.

        Parameters
         - use_all_models (bool=False) Get positions from all models or just the first one.
         - include_alt_loc (bool=False) Get all positions for each atom, or just the first one.
        """
        for model in self.iter_models(use_all_models):
            yield from model.iter_positions(include_alt_loc)

    def __len__(self):
        return len(self.models)

    def _add_atom(self, atom):
        """ """
        if self._current_model is None:
            self._add_model(Model(0))
        atom.model_number = self._current_model.number
        # Atom might be alternate position for existing atom
        self._current_model._add_atom(atom)

    def _finalize(self):
        """Establish first and last residues, atoms, etc."""
        for model in self.models:
            model._finalize()

    def get_unit_cell_lengths(self):
        """Get the lengths of the crystallographic unit cell (may be None)."""
        return self._unit_cell_lengths

    def get_unit_cell_angles(self):
        """Get the angles of the crystallographic unit cell (may be None)."""
        return self._unit_cell_angles


class Model:
    """Model holds one model of a PDB structure.

    NMR structures usually have multiple models.  This represents one
    of them.
    """

    def __init__(self, model_number=1):
        self.number = model_number
        self.chains = []
        self._current_chain = None
        self.chains_by_id = {}
        self.connects = []

    def _add_atom(self, atom):
        """ """
        if len(self.chains) == 0:
            self._add_chain(Chain(atom.chain_id))
        # Create a new chain if the chain id has changed
        if self._current_chain.chain_id != atom.chain_id:
            self._add_chain(Chain(atom.chain_id))
        # Create a new chain after TER record, even if ID is the same
        elif self._current_chain.has_ter_record:
            self._add_chain(Chain(atom.chain_id))
        self._current_chain._add_atom(atom)

    def _add_chain(self, chain):
        self.chains.append(chain)
        self._current_chain = chain
        if chain.chain_id not in self.chains_by_id:
            self.chains_by_id[chain.chain_id] = chain

    def get_chain(self, chain_id):
        return self.chains_by_id[chain_id]

    def chain_ids(self):
        return list(self.chains_by_id.keys())

    def __contains__(self, chain_id):
        return self.chains_by_id.__contains__(chain_id)

    def __getitem__(self, chain_id):
        return self.chains_by_id[chain_id]

    def __iter__(self):
        return iter(self.chains)

    def iter_chains(self):
        yield from self

    def iter_residues(self):
        for chain in self:
            yield from chain.iter_residues()

    def iter_atoms(self):
        for chain in self:
            yield from chain.iter_atoms()

    def iter_positions(self, include_alt_loc=False):
        for chain in self:
            yield from chain.iter_positions(include_alt_loc)

    def __len__(self):
        return len(self.chains)

    def write(self, output_stream=sys.stdout):
        # Start atom serial numbers at 1
        sn = Model.AtomSerialNumber(1)
        for chain in self.chains:
            chain.write(sn, output_stream)

    def _finalize(self):
        for chain in self.chains:
            chain._finalize()

    class AtomSerialNumber:
        """pdb.Model inner class for pass-by-reference incrementable serial number"""

        def __init__(self, val):
            self.val = val

        def increment(self):
            self.val += 1


class Chain:
    def __init__(self, chain_id=" "):
        self.chain_id = chain_id
        self.residues = []
        self.has_ter_record = False
        self._current_residue = None
        self.residues_by_num_icode = {}
        self.residues_by_number = {}

    def _add_atom(self, atom):
        """ """
        # Create a residue if none have been created
        if len(self.residues) == 0:
            self._add_residue(
                Residue(
                    atom.residue_name_with_spaces,
                    atom.residue_number,
                    atom.insertion_code,
                    atom.alternate_location_indicator,
                    atom.segment_id,
                ),
            )
        # Create a residue if the residue information has changed
        elif self._current_residue.number != atom.residue_number:
            self._add_residue(
                Residue(
                    atom.residue_name_with_spaces,
                    atom.residue_number,
                    atom.insertion_code,
                    atom.alternate_location_indicator,
                    atom.segment_id,
                ),
            )
        elif self._current_residue.insertion_code != atom.insertion_code:
            self._add_residue(
                Residue(
                    atom.residue_name_with_spaces,
                    atom.residue_number,
                    atom.insertion_code,
                    atom.alternate_location_indicator,
                    atom.segment_id,
                ),
            )
        elif self._current_residue.name_with_spaces == atom.residue_name_with_spaces:
            # This is a normal case: number, name, and iCode have not changed
            pass
        elif atom.alternate_location_indicator != " ":
            # OK - this is a point mutation, Residue._add_atom will know what to do
            pass
        else:  # Residue name does not match
            # Only residue name does not match
            warnings.warn(
                f"WARNING: two consecutive residues with same number ({atom}, {self._current_residue.atoms[-1]})",
            )
            self._add_residue(
                Residue(
                    atom.residue_name_with_spaces,
                    atom.residue_number,
                    atom.insertion_code,
                    atom.alternate_location_indicator,
                    atom.segment_id,
                ),
            )
        self._current_residue._add_atom(atom)

    def _add_residue(self, residue):
        if len(self.residues) == 0:
            residue.is_first_in_chain = True
        self.residues.append(residue)
        self._current_residue = residue
        key = str(residue.number) + residue.insertion_code
        # only store the first residue with a particular key
        if key not in self.residues_by_num_icode:
            self.residues_by_num_icode[key] = residue
        if residue.number not in self.residues_by_number:
            self.residues_by_number[residue.number] = residue

    def write(self, next_serial_number, output_stream=sys.stdout):
        for residue in self.residues:
            residue.write(next_serial_number, output_stream)
        if self.has_ter_record:
            r = self.residues[-1]
            print(
                "TER   %5d      %3s %1s%4d%1s"
                % (
                    next_serial_number.val,
                    r.name_with_spaces,
                    self.chain_id,
                    r.number % 10000,
                    r.insertion_code,
                ),
                file=output_stream,
            )
            next_serial_number.increment()

    def _add_ter_record(self):
        self.has_ter_record = True
        self._finalize()

    def get_residue(self, residue_number, insertion_code=" "):
        return self.residues_by_num_icode[str(residue_number) + insertion_code]

    def __contains__(self, residue_number):
        return self.residues_by_number.__contains__(residue_number)

    def __getitem__(self, residue_number):
        """Returns the FIRST residue in this chain with a particular residue number"""
        return self.residues_by_number[residue_number]

    def __iter__(self):
        yield from self.residues

    def iter_residues(self):
        yield from self

    def iter_atoms(self):
        for res in self:
            yield from res

    def iter_positions(self, include_alt_loc=False):
        for res in self:
            yield from res.iter_positions(include_alt_loc)

    def __len__(self):
        return len(self.residues)

    def _finalize(self):
        self.residues[0].is_first_in_chain = True
        self.residues[-1].is_final_in_chain = True
        for residue in self.residues:
            residue._finalize()


class Residue:
    def __init__(
        self,
        name,
        number,
        insertion_code=" ",
        primary_alternate_location_indicator=" ",
        segment_id="",
    ):
        alt_loc = primary_alternate_location_indicator
        self.primary_location_id = alt_loc
        self.segment_id = segment_id
        self.locations = {}
        self.locations[alt_loc] = Residue.Location(alt_loc, name)
        self.name_with_spaces = name
        self.number = number
        self.insertion_code = insertion_code
        self.atoms = []
        self.atoms_by_name = {}
        self.is_first_in_chain = False
        self.is_final_in_chain = False
        self._current_atom = None

    def _add_atom(self, atom):
        """ """
        alt_loc = atom.alternate_location_indicator
        if alt_loc not in self.locations:
            self.locations[alt_loc] = Residue.Location(
                alt_loc,
                atom.residue_name_with_spaces,
            )
        assert atom.residue_number == self.number
        assert atom.insertion_code == self.insertion_code

        # Check whether this is an existing atom with another position
        if atom.name_with_spaces in self.atoms_by_name:
            old_atom = self.atoms_by_name[atom.name_with_spaces]
            # Unless this is a duplicated atom (warn about file error)
            if atom.alternate_location_indicator in old_atom.locations:
                pass  # TJL COMMENTED OUT
                # warnings.warn(
                #   "WARNING: duplicate atom (%s, %s)" % (
                #       atom,
                #       old_atom._pdb_string(old_atom.serial_number, atom.alternate_location_indicator),
                #   )
                # )
            else:
                for alt_loc, position in atom.locations.items():
                    old_atom.locations[alt_loc] = position
                return  # no new atom added

        # actually use new atom
        self.atoms_by_name[atom.name] = atom
        self.atoms_by_name[atom.name_with_spaces] = atom
        self.atoms.append(atom)
        self._current_atom = atom

    def write(self, next_serial_number, output_stream=sys.stdout, alt_loc="*"):
        for atom in self.atoms:
            atom.write(next_serial_number, output_stream, alt_loc)

    def _finalize(self):
        if len(self.atoms) > 0:
            self.atoms[0].is_first_atom_in_chain = self.is_first_in_chain
            self.atoms[-1].is_final_atom_in_chain = self.is_final_in_chain
            for atom in self.atoms:
                atom.is_first_residue_in_chain = self.is_first_in_chain
                atom.is_final_residue_in_chain = self.is_final_in_chain

    def set_name_with_spaces(self, name, alt_loc=None):
        # Gromacs ffamber PDB files can have 4-character residue names
        # assert len(name) == 3
        if alt_loc is None:
            alt_loc = self.primary_location_id
        loc = self.locations[alt_loc]
        loc.name_with_spaces = name
        loc.name = name.strip()

    def get_name_with_spaces(self, alt_loc=None):
        if alt_loc is None:
            alt_loc = self.primary_location_id
        loc = self.locations[alt_loc]
        return loc.name_with_spaces

    name_with_spaces = property(
        get_name_with_spaces,
        set_name_with_spaces,
        doc="four-character residue name including spaces",
    )

    def get_name(self, alt_loc=None):
        if alt_loc is None:
            alt_loc = self.primary_location_id
        loc = self.locations[alt_loc]
        return loc.name

    name = property(get_name, doc="residue name")

    def get_atom(self, atom_name):
        return self.atoms_by_name[atom_name]

    def __contains__(self, atom_name):
        return self.atoms_by_name.__contains__(atom_name)

    def __getitem__(self, atom_name):
        """Returns the FIRST atom in this residue with a particular atom name"""
        return self.atoms_by_name[atom_name]

    def __iter__(self):
        "Iterator over atoms"
        yield from self.iter_atoms()

    # Three possibilities: primary alt_loc, certain alt_loc, or all alt_locs
    def iter_atoms(self, alt_loc=None):
        if alt_loc is None:
            locs = [self.primary_location_id]
        elif alt_loc == "":
            locs = [self.primary_location_id]
        elif alt_loc == "*":
            locs = None
        else:
            locs = list(alt_loc)
        # If an atom has any location in alt_loc, emit the atom
        for atom in self.atoms:
            use_atom = False  # start pessimistic
            for loc2 in atom.locations.keys():
                if locs is None:  # means all locations
                    use_atom = True
                elif loc2 in locs:
                    use_atom = True
            if use_atom:
                yield atom

    def iter_positions(self, include_alt_loc=False):
        """Returns one position per atom, even if an individual atom has multiple positions."""
        for atom in self:
            if include_alt_loc:
                yield from atom.iter_positions()
            else:
                yield atom.position

    def __len__(self):
        return len(self.atoms)

    # Residues can have multiple locations, based on alt_loc indicator
    class Location:
        """
        Inner class of residue to allow different residue names for different alternate_locations.
        """

        def __init__(self, alternate_location_indicator, residue_name_with_spaces):
            self.alternate_location_indicator = alternate_location_indicator
            self.residue_name_with_spaces = residue_name_with_spaces


class Atom:
    """Atom represents one atom in a PDB structure."""

    def __init__(self, pdb_line, pdbstructure=None):
        """Create a new pdb.Atom from an ATOM or HETATM line.

        Example line:
        ATOM   2209  CB  TYR A 299       6.167  22.607  20.046  1.00  8.12           C
        00000000011111111112222222222333333333344444444445555555555666666666677777777778
        12345678901234567890123456789012345678901234567890123456789012345678901234567890

        ATOM line format description from
          http://deposit.rcsb.org/adit/docs/pdb_atom_format.html:

        COLUMNS        DATA TYPE       CONTENTS
        --------------------------------------------------------------------------------
         1 -  6        Record name     "ATOM  "
         7 - 11        Integer         Atom serial number.
        13 - 16        Atom            Atom name.
        17             Character       Alternate location indicator.
        18 - 20        Residue name    Residue name.
        22             Character       Chain identifier.
        23 - 26        Integer         Residue sequence number.
        27             AChar           Code for insertion of residues.
        31 - 38        Real(8.3)       Orthogonal coordinates for X in Angstroms.
        39 - 46        Real(8.3)       Orthogonal coordinates for Y in Angstroms.
        47 - 54        Real(8.3)       Orthogonal coordinates for Z in Angstroms.
        55 - 60        Real(6.2)       Occupancy (Default = 1.0).
        61 - 66        Real(6.2)       Temperature factor (Default = 0.0).
        73 - 76        LString(4)      Segment identifier, left-justified.
        77 - 78        LString(2)      Element symbol, right-justified.
        79 - 80        LString(2)      Charge on the atom.

        """
        # We might modify first/final status during _finalize() methods
        self.is_first_atom_in_chain = False
        self.is_final_atom_in_chain = False
        self.is_first_residue_in_chain = False
        self.is_final_residue_in_chain = False
        # Start parsing fields from pdb line
        self.record_name = pdb_line[0:6].strip()

        self.serial_number = _read_atom_number(pdb_line[6:11], pdbstructure=pdbstructure)

        self.name_with_spaces = pdb_line[12:16]
        alternate_location_indicator = pdb_line[16]

        self.residue_name_with_spaces = pdb_line[17:20]
        # In some MD codes, notably ffamber in gromacs, residue name has a fourth character in
        # column 21
        possible_fourth_character = pdb_line[20:21]
        if possible_fourth_character != " ":
            # Fourth character should only be there if official 3 are already full
            if len(self.residue_name_with_spaces.strip()) != 3:
                raise ValueError("Misaligned residue name: %s" % pdb_line)
            self.residue_name_with_spaces += possible_fourth_character
        self.residue_name = self.residue_name_with_spaces.strip()

        self.chain_id = pdb_line[21]
        self.residue_number = _read_residue_number(pdb_line[22:26], pdbstructure, self)

        self.insertion_code = pdb_line[26]
        # coordinates, occupancy, and temperature factor belong in Atom.Location object
        x = float(pdb_line[30:38])
        y = float(pdb_line[38:46])
        z = float(pdb_line[46:54])
        try:
            occupancy = float(pdb_line[54:60])
        except ValueError:
            occupancy = 1.0
        try:
            temperature_factor = float(pdb_line[60:66])
        except ValueError:
            temperature_factor = 0.0
        self.locations = {}
        loc = Atom.Location(
            alternate_location_indicator,
            np.array([x, y, z]),
            occupancy,
            temperature_factor,
            self.residue_name_with_spaces,
        )
        self.locations[alternate_location_indicator] = loc
        self.default_location_id = alternate_location_indicator
        # segment id, element_symbol, and formal_charge are not always present
        self.segment_id = pdb_line[72:76].strip()
        self.element_symbol = pdb_line[76:78].strip()

        # Handle charges
        self.formal_charge = None
        charge_string = pdb_line[78:80].strip()  # Is there any charge information?
        if charge_string:
            # A PDB charge string may either be of format
            # {charge}{sign} or {sign}{charge}
            # We should try both.
            try:
                # Try the string as is. This will work for {sign}{charge}
                self.formal_charge = int(charge_string)
            except ValueError:
                # Try reversing (other format)
                try:
                    charge_string = charge_string[::-1]
                    self.formal_charge = int(charge_string)
                except ValueError:
                    warnings.warn(f"Could not parse charge information for atom {self._pdb_string()}\nSetting to None")
        # figure out atom element
        try:
            # First try to find a sensible element symbol from columns 76-77
            self.element = element.get_by_symbol(self.element_symbol)
        except KeyError:
            # otherwise, deduce element from first two characters of atom name
            # remove digits found in some hydrogen atom names
            symbol = self.name_with_spaces[0:2].strip().lstrip("0123456789")
            try:
                # Some molecular dynamics PDB files, such as gromacs with ffamber force
                # field, include 4-character hydrogen atom names beginning with "H".
                # Hopefully elements like holmium (Ho) and mercury (Hg) will have fewer than four
                # characters in the atom name.  This problem is the fault of molecular
                # dynamics code authors who feel the need to make up their own atom
                # nomenclature because it is too tedious to read that provided by the PDB.
                # These are the same folks who invent their own meanings for biochemical terms
                # like "dipeptide".  Clowntards.
                if len(self.name) == 4 and self.name[0:1] == "H":
                    self.element = element.hydrogen
                else:
                    self.element = element.get_by_symbol(symbol)
            except KeyError:
                # OK, I give up
                self.element = None
        if pdbstructure is not None:
            pdbstructure._next_atom_number = self.serial_number + 1
            pdbstructure._next_residue_number = self.residue_number + 1

    def iter_locations(self):
        """
        Iterate over Atom.Location objects for this atom, including primary location.
        """
        for alt_loc in self.locations:
            yield self.locations[alt_loc]

    def iter_positions(self):
        """
        Iterate over atomic positions.  Returns Quantity(Vec3(), unit) objects, unlike
        iter_locations, which returns Atom.Location objects.
        """
        for loc in self.iter_locations():
            yield loc.position

    def iter_coordinates(self):
        """
        Iterate over x, y, z values of primary atom position.
        """
        yield from self.position

    # Hide existence of multiple alternate locations to avoid scaring casual users
    def get_location(self, location_id=None):
        id = location_id
        if id is None:
            id = self.default_location_id
        return self.locations[id]

    def set_location(self, new_location, location_id=None):
        id = location_id
        if id is None:
            id = self.default_location_id
        self.locations[id] = new_location

    location = property(get_location, set_location, doc="default Atom.Location object")

    def get_position(self):
        return self.location.position

    def set_position(self, coords):
        self.location.position = coords

    position = property(get_position, set_position, doc="orthogonal coordinates")

    def get_alternate_location_indicator(self):
        return self.location.alternate_location_indicator

    alternate_location_indicator = property(get_alternate_location_indicator)

    def get_occupancy(self):
        return self.location.occupancy

    occupancy = property(get_occupancy)

    def get_temperature_factor(self):
        return self.location.temperature_factor

    temperature_factor = property(get_temperature_factor)

    def get_x(self):
        return self.position[0]

    x = property(get_x)

    def get_y(self):
        return self.position[1]

    y = property(get_y)

    def get_z(self):
        return self.position[2]

    z = property(get_z)

    def _pdb_string(self, serial_number=None, alternate_location_indicator=None):
        """
        Produce a PDB line for this atom using a particular serial number and alternate location
        """
        if serial_number is None:
            serial_number = self.serial_number
        if alternate_location_indicator is None:
            alternate_location_indicator = self.alternate_location_indicator
        # produce PDB line in three parts: names, numbers, and end
        # Accomodate 4-character residue names that use column 21
        long_res_name = self.residue_name_with_spaces
        if len(long_res_name) == 3:
            long_res_name += " "
        assert len(long_res_name) == 4
        names = "%-6s%5d %4s%1s%4s%1s%4d%1s   " % (
            self.record_name,
            serial_number,
            self.name_with_spaces,
            alternate_location_indicator,
            long_res_name,
            self.chain_id,
            self.residue_number,
            self.insertion_code,
        )
        numbers = f"{self.x:8.3f}{self.y:8.3f}{self.z:8.3f}{self.occupancy:6.2f}{self.temperature_factor:6.2f}      "
        end = "%-4s%2s" % (
            self.segment_id,
            self.element_symbol,
        )
        formal_charge = "  "
        if self.formal_charge is not None:
            formal_charge = "%+2d" % self.formal_charge
        return names + numbers + end + formal_charge

    def __str__(self):
        return self._pdb_string(self.serial_number, self.alternate_location_indicator)

    def write(self, next_serial_number, output_stream=sys.stdout, alt_loc="*"):
        """
        alt_loc = "*" means write all alternate locations
        alt_loc = None means write just the primary location
        alt_loc = "AB" means write locations "A" and "B"
        """
        if alt_loc is None:
            locs = [self.default_location_id]
        elif alt_loc == "":
            locs = [self.default_location_id]
        elif alt_loc == "*":
            locs = self.locations.keys()
            locs.sort()
        else:
            locs = list(alt_loc)
        for loc_id in locs:
            print(self._pdb_string(next_serial_number.val, loc_id), file=output_stream)
            next_serial_number.increment()

    def set_name_with_spaces(self, name):
        assert len(name) == 4
        self._name_with_spaces = name
        self._name = name.strip()

    def get_name_with_spaces(self):
        return self._name_with_spaces

    name_with_spaces = property(
        get_name_with_spaces,
        set_name_with_spaces,
        doc="four-character residue name including spaces",
    )

    def get_name(self):
        return self._name

    name = property(get_name, doc="residue name")

    class Location:
        """
        Inner class of Atom for holding alternate locations
        """

        def __init__(
            self,
            alt_loc,
            position,
            occupancy,
            temperature_factor,
            residue_name,
        ):
            self.alternate_location_indicator = alt_loc
            self.position = position
            self.occupancy = occupancy
            self.temperature_factor = temperature_factor
            self.residue_name = residue_name

        def __iter__(self):
            yield from self.position

        def __str__(self):
            return str(self.position)



# #############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
# Molecular Dynamics Trajectories.
# Copyright 2012-2014 Stanford University and the Authors
#
# Authors: Matthew Harrigan
# Contributors: Robert T. McGibbon
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
# #############################################################################


import ast
import re
import sys
from ast import unparse
from collections import namedtuple
from copy import deepcopy

from pyparsing import (
    Group,
    Keyword,
    MatchFirst,
    OneOrMore,
    OpAssoc,
    ParseException,
    ParserElement,
    Word,
    alphanums,
    alphas,
    infix_notation,
    quoted_string,
)

# this number arises from the current selection language, if the cache size is exceeded, it hurts performance a bit.
ParserElement.enable_packrat(cache_size_limit=304)

__all__ = ["parse_selection"]

# ############################################################################
# Globals
# ############################################################################

NUMS = ".0123456789"
THIS_ATOM = ast.Name(id="atom", ctx=ast.Load())
RE_MODULE = ast.Name(id="re", ctx=ast.Load())
SINGLETON_NODE_IDS = {THIS_ATOM.id, RE_MODULE.id}

SELECTION_GLOBALS = {"re": re}
_ParsedSelection = namedtuple("_ParsedSelection", ["expr", "source", "astnode"])

# ############################################################################
# Utils
# ############################################################################


class _RewriteNames(ast.NodeTransformer):
    def visit_Name(self, node):
        if node.id in SINGLETON_NODE_IDS:
            return node

        _safe_names = {"None": None, "True": True, "False": False}
        if node.id in _safe_names:
            return ast.Constant(value=_safe_names[node.id], kind=None)

        # all other bare names are taken to be string literals. Thus something
        # like parse_selection('name CA') properly resolves CA as a string
        # literal, not a barename to be loaded from the global scope!
        return ast.Constant(value=node.id, kind=None)


def _chain(*attrs):
    """This transforms, for example, ('residue', 'is_protein'), into
    Attribute(value=Attribute(value=THIS_ATOM,
     attr='residue', ctx=Load()), attr='is_protein', ctx=Load())
    """
    left = THIS_ATOM
    for attr in attrs:
        left = ast.Attribute(value=left, attr=attr, ctx=ast.Load())
    return left


def _kw(*tuples):
    """Create a many-to-one dictionary.

    _kw((['one', '1'], 'one'))
    gives {'one': 'one', '1': 'one'}
    """
    dic = dict()
    for keys, val in tuples:
        for key in keys:
            dic[key] = val
    return dic


def _check_n_tokens(tokens, n_tokens, name):
    if not len(tokens) == n_tokens:
        err = "{} take {} values. You gave {}"
        err = err.format(name, n_tokens, len(tokens))
        raise ParseException(err)


class SelectionKeyword:
    keyword_aliases = _kw(
        # Atom.<attribute>
        (("all", "everything"), ast.Name(id="True", ctx=ast.Load())),
        (("none", "nothing"), ast.Name(id="False", ctx=ast.Load())),
        (("backbone", "is_backbone"), _chain("is_backbone")),
        (("sidechain", "is_sidechain"), _chain("is_sidechain")),
        # Atom.residue.<attribute>
        (("protein", "is_protein"), _chain("residue", "is_protein")),
        (("code", "rescode", "resc"), _chain("residue", "code")),
        (("nucleic", "is_nucleic"), _chain("residue", "is_nucleic")),
        (("water", "waters", "is_water"), _chain("residue", "is_water")),
        (("name",), _chain("name")),
        (("index",), _chain("index")),
        (("n_bonds",), _chain("n_bonds")),
        (("residue", "resSeq"), _chain("residue", "resSeq")),
        (("resname", "resn"), _chain("residue", "name")),
        (("resid", "resi"), _chain("residue", "index")),
        (("segment_id", "segname"), _chain("segment_id")),
        # Atom.residue.chain.<attribute>
        (("chainid",), _chain("residue", "chain", "index")),
        # Atom.element.<attribute>
        (("type", "element", "symbol"), _chain("element", "symbol")),
        # (('radius',), _chain('element', 'radius')),
        (("mass",), _chain("element", "mass")),
    )

    def __init__(self, tokens):
        # pyparsing constructs the instance while building the parse tree,
        # and gives us the set tokens. In this case, the tokens are
        self._tokens = tokens
        _check_n_tokens(tokens, 1, "Unary selectors")
        assert tokens[0] in self.keyword_aliases

    def ast(self):
        return self.keyword_aliases[self._tokens[0]]


class Literal:
    def __init__(self, tokens):
        self.token = tokens[0]
        _check_n_tokens(tokens, 1, "literal")

    def ast(self):
        return ast.parse(self.token, mode="eval").body


class UnaryInfixOperand:
    n_terms = 1
    assoc = "RIGHT"

    keyword_aliases = _kw(
        (["not ", "!"], ast.Not()),
    )

    def __init__(self, tokens):
        tokens = tokens[0]
        _check_n_tokens(tokens, 2, "Unary infix operators")
        self.op_token, self.value_token = tokens
        assert self.op_token in self.keyword_aliases
        if isinstance(self.value_token, Literal):
            raise ValueError("Cannot use literals as booleans.")

    def ast(self):
        return ast.UnaryOp(
            op=self.keyword_aliases[self.op_token],
            operand=self.value_token.ast(),
        )


class RegexInfixOperand:
    n_terms = 2
    assoc = "LEFT"
    keyword_aliases = {"=~": "=~"}

    def __init__(self, tokens):
        self.tokens = tokens[0]
        _check_n_tokens(self.tokens, 3, "regex operator")
        self.string, op, self.pattern = self.tokens
        assert op == "=~"
        if isinstance(self.string, Literal):
            raise ValueError("Cannot do regex comparison on literal")

    def ast(self):
        pattern = self.tokens[2].ast()
        string = self.tokens[0].ast()
        return ast.Compare(
            left=ast.Call(
                func=ast.Attribute(
                    value=RE_MODULE,
                    attr="match",
                    ctx=ast.Load(),
                ),
                args=[pattern, string],
                keywords=[],
            ),
            ops=[ast.IsNot()],
            comparators=[ast.Name(id="None", ctx=ast.Load())],
        )


class BinaryInfixOperand:
    n_terms = 2
    assoc = "LEFT"

    keyword_aliases = _kw(
        (["and", "&&"], ast.And()),
        (["or", "||"], ast.Or()),
        (["<", "lt"], ast.Lt()),
        (["==", "eq"], ast.Eq()),
        (["<=", "le"], ast.LtE()),
        (["!=", "ne"], ast.NotEq()),
        ([">=", "ge"], ast.GtE()),
        ([">", "gt"], ast.Gt()),
    )

    def __init__(self, tokens):
        tokens = tokens[0]
        if len(tokens) % 2 == 1:
            self.op_token = tokens[1]
            self.comparators = tokens[::2]
        else:
            err = "Invalid number of infix expressions: {}"
            err = err.format(len(tokens))
            raise ParseException(err)
        assert self.op_token in self.keyword_aliases

        # Check for too many literals and not enough keywords
        op = self.keyword_aliases[self.op_token]
        if isinstance(op, ast.boolop):
            if any(isinstance(c, Literal) for c in self.comparators):
                raise ValueError("Cannot use literals as truth")
        else:
            if all(isinstance(c, Literal) for c in self.comparators):
                raise ValueError("Cannot compare literals.")

    def ast(self):
        op = self.keyword_aliases[self.op_token]

        if isinstance(op, ast.boolop):
            # and and or use one type of AST node
            value = ast.BoolOp(op=op, values=[e.ast() for e in self.comparators])
        else:
            # remaining operators use another
            value = ast.Compare(
                left=self.comparators[0].ast(),
                ops=[op],
                comparators=[e.ast() for e in self.comparators[1:]],
            )
        return value


class RangeCondition:
    def __init__(self, tokens):
        tokens = tokens[0]
        _check_n_tokens(tokens, 4, "range condition")
        assert tokens[2] == "to"
        self._field, self._from, self._to = tokens[0], tokens[1], tokens[3]
        if isinstance(self._field, Literal):
            raise ValueError("Can't test literal in range.")

    def ast(self):
        return ast.Compare(
            left=self._from.ast(),
            ops=[ast.LtE(), ast.LtE()],
            comparators=[self._field.ast(), self._to.ast()],
        )


class InListCondition:
    def __init__(self, tokens):
        tokens = tokens[0]
        self._field = tokens[0]

        if len(tokens) == 2:
            # Implicit equality
            self.implicit_equality = True
        elif len(tokens) > 2:
            self.implicit_equality = False
        else:
            raise ValueError("Not enough tokens for `in` condition")

        self.compare_to = tokens[1:]

    def ast(self):
        if self.implicit_equality:
            return ast.Compare(
                left=self._field.ast(),
                ops=[ast.Eq()],
                comparators=[e.ast() for e in self.compare_to],
            )
        else:
            comparator = ast.List([e.ast() for e in self.compare_to], ast.Load())
            return ast.Compare(
                left=self._field.ast(),
                ops=[ast.In()],
                comparators=[comparator],
            )


class parse_selection:
    """Parse an atom selection expression

    Parameters
    ----------
    selection_string : str
        Selection string, a string in the MDTraj atom selection grammar.

    Returns
    -------
    expr : callable (atom -> bool)
        A callable object which accepts an MDTraj.core.topology.Atom object and
        returns a boolean value giving whether or not that particular atom
        satisfies the selection string.
    source : str
        Python source code corresponding to the expression ``expr``.
    astnode : ast.AST
        Python abstract syntax tree node containing the parsed expression

    Examples
    --------
    >>> expr, source, astnode = parse_selection('protein and type CA')
    >>> expr
    <function __main__.<lambda>>
    >>> source
    '(atom.residue.is_protein and (atom.element.symbol == CA))'
    >>> <_ast.BoolOp at 0x103969d50>
    """

    def __init__(self):
        self.is_initialized = False
        self.expression = None

    def _initialize(self):
        def keywords(klass):
            kws = sorted(klass.keyword_aliases.keys())
            return MatchFirst([Keyword(kw) for kw in kws])

        def infix(klass):
            kws = sorted(klass.keyword_aliases.keys())
            return [(kw, klass.n_terms, getattr(OpAssoc, klass.assoc), klass) for kw in kws]

        # literals include words made of alphanumerics, numbers,
        # or quoted strings but we exclude any of the logical
        # operands (e.g. 'or') from being parsed literals
        literal = ~(keywords(BinaryInfixOperand) | keywords(UnaryInfixOperand)) + (
            Word(NUMS) | quoted_string | Word(alphas, alphanums)
        )
        literal.set_parse_action(Literal)

        # These are the other 'root' expressions,
        # the selection keywords (resname, resid, mass, etc)
        selection_keyword = keywords(SelectionKeyword)
        selection_keyword.set_parse_action(SelectionKeyword)
        base_expression = MatchFirst([selection_keyword, literal])

        # range condition matches expressions such as 'mass 1 to 20'
        range_condition = Group(
            selection_keyword + literal + Keyword("to") + literal,
        )
        range_condition.set_parse_action(RangeCondition)

        # matches expression such as `resname GLU ASP ARG` and also
        # handles implicit equality `resname ALA`
        in_list_condition = Group(selection_keyword + OneOrMore(literal))
        in_list_condition.set_parse_action(InListCondition)

        expression = range_condition | in_list_condition | base_expression
        logical_expr = infix_notation(
            expression,
            infix(UnaryInfixOperand) + infix(BinaryInfixOperand) + infix(RegexInfixOperand),
        )

        self.expression = logical_expr
        self.is_initialized = True

        self.transformer = _RewriteNames()

    def __call__(self, selection):
        if not self.is_initialized:
            self._initialize()

        try:
            parse_result = self.expression.parse_string(selection, parse_all=True)
        except ParseException as e:
            msg = str(e)
            lines = [
                f"{msg}: {selection}",
                " " * (12 + len("%s: " % msg) + e.loc) + "^^^",
            ]
            raise ValueError("\n".join(lines))

        # Change __ATOM__ in function bodies. It must bind to the arg
        # name specified below (i.e. 'atom')
        astnode = self.transformer.visit(deepcopy(parse_result[0].ast()))

        # Special check for a single literal

        if isinstance(astnode, ast.Constant) and astnode.value not in {True, False, None}:
            raise ValueError(
                f"Cannot use a single literal as a boolean. Choked on node with value {astnode.value}",
            )

        args = [ast.arg(arg="atom", annotation=None)]
        signature = ast.arguments(
            args=args,
            vararg=None,
            kwarg=None,
            posonlyargs=[],
            kwonlyargs=[],
            defaults=[],
            kw_defaults=[],
        )

        func = ast.Expression(body=ast.Lambda(signature, astnode))
        source = unparse(astnode)

        expr = eval(
            compile(ast.fix_missing_locations(func), "<string>", mode="eval"),
            SELECTION_GLOBALS,
        )
        return _ParsedSelection(expr, source, astnode)


# Create the callable, and use it to overshadow the class. this way there's
# basically just one global instance of the "function", even thought its
# a callable class.
parse_selection = parse_selection()

if __name__ == "__main__":
    exp = parse_selection(sys.argv[1])
    print(exp.source)
    print(ast.dump(exp.astnode))



from .scalar_couplings import compute_J3_HN_C, compute_J3_HN_CB, compute_J3_HN_HA
from .shift_wrappers import (
    chemical_shifts_ppm,
    chemical_shifts_shiftx2,
    chemical_shifts_spartaplus,
    compute_chemical_shifts,
    reindex_dataframe_by_atoms,
)

__all__ = (
    "compute_J3_HN_C",
    "compute_J3_HN_CB",
    "compute_J3_HN_HA",
    "chemical_shifts_ppm",
    "chemical_shifts_shiftx2",
    "chemical_shifts_spartaplus",
    "compute_chemical_shifts",
    "reindex_dataframe_by_atoms",
)



##
# File:  PdbxReader.py
# Date:  2012-01-09  Jdw  Adapted from PdbxParser
#
# Updates:
#
# 2012-01-09 - (jdw) Separate reader and writer classes.
#
# 2012-09-02 - (jdw)  Revise tokenizer to better handle embedded quoting.
#
# 2025-05-15 - (cb) Changes docstrings to mdtraj format, exception handling.
#                   Replaces import directory from OpenMM.
#                   Removes unused __tokenizerOrg
##
"""
PDBx/mmCIF dictionary and data file parser.

Acknowledgements
---------------
The tokenizer used in this module is modeled after the clever parser design
used in the PyMMLIB package.

PyMMLib Development Group
Authors: Ethan Merritt (merritt@u.washington.ed) & Jay Painter (jay.painter@gmail.com)
See: http://pymmlib.sourceforge.net/
"""

import re

from .PdbxContainers import DataCategory, DataContainer, DefinitionContainer


class PdbxError(Exception):
    """
    Class for catch general errors
    """

    pass


class SyntaxError(Exception):
    """
    Exception raised for syntax errors in the mmCIF file.

    Parameters
    ----------
    lineNumber : int
        The line number where the error occurred.
    text : str
        Description of the syntax error.
    """

    def __init__(self, lineNumber, text):
        super().__init__()
        self.lineNumber = lineNumber
        self.text = text

    def __str__(self):
        return f"%ERROR - [at line: {self.lineNumber}] {self.text}"


class PdbxReader:
    """
    PDBx reader for data files and dictionaries.

    Parameters
    ----------
    ifh : file-like object
        Input file handle returned by open().
    """

    def __init__(self, ifh):
        """ifh - input file handle returned by open()"""
        #
        self.__curLineNumber = 0
        self.__ifh = ifh
        self.__stateDict = {
            "data": "ST_DATA_CONTAINER",
            "loop": "ST_TABLE",
            "global": "ST_GLOBAL_CONTAINER",
            "save": "ST_DEFINITION",
            "stop": "ST_STOP",
        }

    def read(self, containerList):
        """
        Parse the input file and append data and definition containers to containerList.

        Parameters
        ----------
        containerList : list
            List to which the parsed data and definition objects will be appended.

        Raises
        ------
        PdbxError
            If an unexpected error occurs during parsing.
        """
        self.__curLineNumber = 0
        try:
            self.__parser(self.__tokenizer(self.__ifh), containerList)
        except StopIteration:
            pass
        except RuntimeError as err:
            if "StopIteration" not in str(err):
                raise
        else:
            raise PdbxError()

    def __syntaxError(self, errText):
        raise SyntaxError(self.__curLineNumber, errText)

    def __getContainerName(self, inWord):
        """
        Returns the name of the data_ or save_ container

        Parameters
        ----------
        inWord : str
            A reserved word starting with 'data_' or 'save_'.

        Returns
        -------
        str
            The container name.
        """
        return str(inWord[5:]).strip()

    def __getState(self, inWord):
        """
        Identifies reserved syntax elements and assigns an associated state.

        Parameters
        ----------
        inWord : str
            Token representing a reserved syntax element.

        Returns
        -------
        tuple
            A tuple (reserved_word, state) where reserved_word is one of the CIF syntax elements
            (e.g., data_, loop_, global_, save_, stop_) and state is the corresponding parser state
            required to process the next section.
        """
        i = inWord.find("_")
        if i == -1:
            return None, "ST_UNKNOWN"

        try:
            rWord = inWord[:i].lower()
            return rWord, self.__stateDict[rWord]
        except KeyError:
            return None, "ST_UNKNOWN"

    def __parser(self, tokenizer, containerList):
        """
        Parse PDBx data files and dictionaries.

        This method iterates over the tokens from the tokenizer and populates the
        containerList with data and definition objects.

        Parameters
        ----------
        tokenizer : generator
            A reentrant method recognizing data item names (_category.attribute)
            quoted strings (single, double and multi-line semi-colon delimited),
            and unquoted strings.
        containerList : list
            Container for data and definition objects parsed from the input file.

        Raises
        ------
        SyntaxError
            If any syntax error is encountered during parsing.
        """
        # Working container - data or definition
        curContainer = None
        #
        # Working category container
        categoryIndex = {}
        curCategory = None
        #
        curRow = None
        state = None

        # Find the first reserved word and begin capturing data.
        #
        while True:
            curCatName, curAttName, curQuotedString, curWord = next(tokenizer)
            if curWord is None:
                continue
            reservedWord, state = self.__getState(curWord)
            if reservedWord is not None:
                break

        while True:
            #
            #  Set the current state  -
            #
            #  At this point in the processing cycle we are expecting a token containing
            #  either a '_category.attribute'  or a reserved word.
            #
            if curCatName is not None:
                state = "ST_KEY_VALUE_PAIR"
            elif curWord is not None:
                reservedWord, state = self.__getState(curWord)
            else:
                self.__syntaxError("Miscellaneous syntax error")
                return

            #
            # Process  _category.attribute  value assignments
            #
            if state == "ST_KEY_VALUE_PAIR":
                try:
                    curCategory = categoryIndex[curCatName]
                except KeyError:
                    # A new category is encountered - create a container and add a row
                    curCategory = categoryIndex[curCatName] = DataCategory(curCatName)

                    try:
                        curContainer.append(curCategory)
                    except AttributeError:
                        self.__syntaxError("Category cannot be added to  data_ block")
                        return

                    curRow = []
                    curCategory.append(curRow)
                else:
                    # Recover the existing row from the category
                    try:
                        curRow = curCategory[0]
                    except IndexError:
                        self.__syntaxError(
                            "Internal index error accessing category data",
                        )
                        return

                # Check for duplicate attributes and add attribute to table.
                if curAttName in curCategory.getAttributeList():
                    self.__syntaxError("Duplicate attribute encountered in category")
                    return
                else:
                    curCategory.appendAttribute(curAttName)

                # Get the data for this attribute from the next token
                tCat, tAtt, curQuotedString, curWord = next(tokenizer)

                if tCat is not None or (curQuotedString is None and curWord is None):
                    self.__syntaxError(
                        f"Missing data for item _{curCatName}.{curAttName}",
                    )

                if curWord is not None:
                    #
                    # Validation check token for misplaced reserved words  -
                    #
                    reservedWord, state = self.__getState(curWord)
                    if reservedWord is not None:
                        self.__syntaxError(
                            f"Unexpected reserved word: {reservedWord}",
                        )

                    curRow.append(curWord)

                elif curQuotedString is not None:
                    curRow.append(curQuotedString)

                else:
                    self.__syntaxError("Missing value in item-value pair")

                curCatName, curAttName, curQuotedString, curWord = next(tokenizer)
                continue

            #
            # Process a loop_ declaration and associated data -
            #
            elif state == "ST_TABLE":
                # The category name in the next curCatName,curAttName pair
                #    defines the name of the category container.
                curCatName, curAttName, curQuotedString, curWord = next(tokenizer)

                if curCatName is None or curAttName is None:
                    self.__syntaxError("Unexpected token in loop_ declaration")
                    return

                # Check for a previous category declaration.
                if curCatName in categoryIndex:
                    self.__syntaxError("Duplicate category declaration in loop_")
                    return

                curCategory = DataCategory(curCatName)

                try:
                    curContainer.append(curCategory)
                except AttributeError:
                    self.__syntaxError(
                        "loop_ declaration outside of data_ block or save_ frame",
                    )
                    return

                curCategory.appendAttribute(curAttName)

                # Read the rest of the loop_ declaration
                while True:
                    curCatName, curAttName, curQuotedString, curWord = next(tokenizer)

                    if curCatName is None:
                        break

                    if curCatName != curCategory.getName():
                        self.__syntaxError("Changed category name in loop_ declaration")
                        return

                    curCategory.appendAttribute(curAttName)

                # If the next token is a 'word', check it for any reserved words -
                if curWord is not None:
                    reservedWord, state = self.__getState(curWord)
                    if reservedWord is not None:
                        if reservedWord == "stop":
                            return
                        else:
                            self.__syntaxError(
                                f"Unexpected reserved word after loop declaration: {reservedWord}",
                            )

                # Read the table of data for this loop_ -
                while True:
                    curRow = []
                    curCategory.append(curRow)

                    for tAtt in curCategory.getAttributeList():
                        if curWord is not None:
                            curRow.append(curWord)
                        elif curQuotedString is not None:
                            curRow.append(curQuotedString)

                        curCatName, curAttName, curQuotedString, curWord = next(
                            tokenizer,
                        )

                    # loop_ data processing ends if -

                    # A new _category.attribute is encountered
                    if curCatName is not None:
                        break

                    # A reserved word is encountered
                    if curWord is not None:
                        reservedWord, state = self.__getState(curWord)
                        if reservedWord is not None:
                            break

                continue

            elif state == "ST_DEFINITION":
                # Ignore trailing unnamed saveframe delimiters e.g. 'save_'
                sName = self.__getContainerName(curWord)
                if len(sName) > 0:
                    curContainer = DefinitionContainer(sName)
                    containerList.append(curContainer)
                    categoryIndex = {}
                    curCategory = None

                curCatName, curAttName, curQuotedString, curWord = next(tokenizer)

            elif state == "ST_DATA_CONTAINER":
                #
                dName = self.__getContainerName(curWord)
                if len(dName) == 0:
                    dName = "unidentified"
                curContainer = DataContainer(dName)
                containerList.append(curContainer)
                categoryIndex = {}
                curCategory = None
                curCatName, curAttName, curQuotedString, curWord = next(tokenizer)

            elif state == "ST_STOP":
                return
            elif state == "ST_GLOBAL":
                curContainer = DataContainer("blank-global")
                curContainer.setGlobal()
                containerList.append(curContainer)
                categoryIndex = {}
                curCategory = None
                curCatName, curAttName, curQuotedString, curWord = next(tokenizer)

            elif state == "ST_UNKNOWN":
                self.__syntaxError(f"Unrecogized syntax element: {str(curWord)}")
                return

    def __tokenizer(self, ifh):
        """
        Tokenize the mmCIF file.

        This generator method yields tokens from the input file. Each yielded tuple
        has the following structure:
            (category name, attribute name, quoted string, unquoted word)

        Parameters
        ----------
        ifh : file-like object
            The input file handle.

        Yields
        ------
        tuple
            A tuple containing:
            - category name : str or None
            - attribute name : str or None
            - quoted string : str or None
            - unquoted word : str or None
        """
        #
        # Regex definition for mmCIF syntax - semi-colon delimited strings are handled
        #                                     outside of this regex.
        mmcifRe = re.compile(
            r"(?:"
            r"(?:_(.+?)[.](\S+))"
            "|"  # _category.attribute
            r"(?:['](.*?)(?:[']\s|[']$))"
            "|"  # single quoted strings
            r"(?:[\"](.*?)(?:[\"]\s|[\"]$))"
            "|"  # double quoted strings
            r"(?:\s*#.*$)"
            "|"  # comments (dumped)
            r"(\S+)"  # unquoted words
            r")",
        )

        fileIter = iter(ifh)

        ## Tokenizer loop begins here ---
        while True:
            line = next(fileIter)
            self.__curLineNumber += 1

            # Dump comments
            if line.startswith("#"):
                continue

            # Gobble up the entire semi-colon/multi-line delimited string and
            #    and stuff this into the string slot in the return tuple
            #
            if line.startswith(";"):
                mlString = [line[1:]]
                while True:
                    line = next(fileIter)
                    self.__curLineNumber += 1
                    if line.startswith(";"):
                        break
                    mlString.append(line)

                # remove trailing new-line that is part of the \n; delimiter
                mlString[-1] = mlString[-1].rstrip()
                #
                yield (None, None, "".join(mlString), None)
                #
                # Need to process the remainder of the current line -
                line = line[1:]
                # continue

            # Apply regex to the current line consolidate the single/double
            # quoted within the quoted string category
            for it in mmcifRe.finditer(line):
                tgroups = it.groups()
                if tgroups != (None, None, None, None, None):
                    if tgroups[2] is not None:
                        qs = tgroups[2]
                    elif tgroups[3] is not None:
                        qs = tgroups[3]
                    else:
                        qs = None
                    groups = (tgroups[0], tgroups[1], qs, tgroups[4])
                    yield groups



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2014 Stanford University and the Authors
#
# Authors: Kyle A. Beauchamp
# Contributors: Robert McGibbon, Matthew Harrigan, Carlos Xavier Hernandez
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import os
import pickle
import tempfile

import numpy as np
import pytest

import mdtraj as md
from mdtraj.testing import eq
from mdtraj.utils import import_

try:
    import openmm.unit as u
    from openmm import app

    HAVE_OPENMM = True
except ImportError:
    HAVE_OPENMM = False

needs_openmm = pytest.mark.skipif(not HAVE_OPENMM, reason="needs OpenMM")


def test_topology_amber_nonstandard_names(get_fn):
    # Test that nonstandard protein atom and residue names
    # in AMBER prmtop files can be preserved if desired
    topology = md.load(
        get_fn("issue_2056.inpcrd"),
        top=get_fn("issue_2056.prmtop"),
    ).topology
    assert topology.atom(1).name == "H"
    assert topology.residue(1).name == "HIS"
    topology = md.load(
        get_fn("issue_2056.inpcrd"),
        top=get_fn("issue_2056.prmtop"),
        standard_names=False,
    ).topology
    assert topology.atom(1).name == "H1"
    assert topology.residue(1).name == "HID"


@needs_openmm
def test_topology_openmm(get_fn):
    topology = md.load(get_fn("1bpi.pdb")).topology
    topology_with_bond_order = md.load(get_fn("imatinib.mol2")).topology

    # the openmm trajectory doesn't have the distinction
    # between resSeq and index, so if they're out of whack
    # in the openmm version, that cant be preserved
    for top in [topology, topology_with_bond_order]:
        for residue in top.residues:
            residue.resSeq = residue.index
        mm = top.to_openmm()
        assert isinstance(mm, app.Topology)
        topology2 = md.Topology.from_openmm(mm)
        eq(top, topology2)


@needs_openmm
def test_topology_openmm_preserve_chain_id_when_specified(get_fn):
    top = md.load(get_fn("custom.pdb")).topology
    for chain in top.chains:
        assert chain.chain_id is not None
    openmm_top = top.to_openmm()
    chain = next(openmm_top.chains())
    eq(chain.id, "X")


@needs_openmm
def test_topology_openmm_roundtrip_chain_id(get_fn):
    top = md.load(get_fn("custom.pdb")).topology
    top2 = md.Topology.from_openmm(top.to_openmm())
    for chain1, chain2 in zip(top.chains, top2.chains):
        eq(chain1.chain_id, "X")
        eq(chain2.chain_id, "X")


@needs_openmm
def test_topology_openmm_preserve_chain_id_even_when_empty(get_fn):
    top = md.load(get_fn("native.pdb")).topology
    for chain in top.chains:
        assert chain.chain_id is not None
    openmm_top = top.to_openmm()
    chain = next(openmm_top.chains())
    eq(chain.id, " ")


@needs_openmm
def test_topology_openmm_boxes(get_fn):
    traj = md.load(get_fn("1vii_sustiva_water.pdb"))
    mmtop = traj.topology.to_openmm(traj=traj)
    mmtop.getUnitCellDimensions() / u.nanometer


@needs_openmm
def test_topology_openmm_formal_charges(get_fn):
    """Test to make sure charges are conserved when converting to/from openmm from PDB"""
    app = import_("openmm.app")

    #  This is a edited 1ply PDB with formal charges added to sodium ions
    pdb = app.PDBFile(get_fn("1ply_charge.pdb"))

    # Get the formal charges from the OpenMM topology
    try:
        formal_charges = [atom.formalCharge for atom in pdb.topology.atoms()]
    except TypeError:
        # OpenMM < 7.4 doesn't have formal charges,
        # so we just return (skip the test)
        return

    # Convert to MDTraj topology
    mdtraj_topology = md.Topology.from_openmm(pdb.topology)

    # Get the formal charges from the MDTraj topology
    mdtraj_formal_charges = [atom.formal_charge for atom in mdtraj_topology.atoms]

    # Check that the formal charges are the same
    eq(formal_charges, mdtraj_formal_charges)


def test_topology_dataframe_formal_charges(get_fn):
    """
    Test that formal charges are maintained when converting a topology
    to a DataFrame and back. Uses a PDB file with formal charges.
    """
    # Load the topology from a PDB file with formal charges.
    topology = md.load(get_fn("1ply_charge.pdb")).topology

    # Get the original formal charges from the MDTraj topology.
    original_formal_charges = [atom.formal_charge for atom in topology.atoms]

    # Convert topology to DataFrame and bonds array.
    atoms_df, bonds = topology.to_dataframe()

    # Convert topology back from the DataFrame.
    topology_from_df = md.Topology.from_dataframe(atoms_df, bonds)

    # Extract formal charges from the converted topology.
    converted_formal_charges = [atom.formal_charge for atom in topology_from_df.atoms]

    # Check that formal charges are conserved.
    eq(original_formal_charges, converted_formal_charges)


def test_topology_pandas(get_fn):
    topology = md.load(get_fn("native.pdb")).topology
    atoms, bonds = topology.to_dataframe()

    topology2 = md.Topology.from_dataframe(atoms, bonds)
    eq(topology, topology2)

    # Make sure default argument of None works, see issue #774
    md.Topology.from_dataframe(atoms)


def test_topology_pandas_TIP4PEW(get_fn):
    topology = md.load(get_fn("GG-tip4pew.pdb")).topology
    atoms, bonds = topology.to_dataframe()

    topology2 = md.Topology.from_dataframe(atoms, bonds)
    eq(topology, topology2)


def test_topology_pandas_2residues_same_resSeq(get_fn):
    with pytest.warns(UserWarning, match="two consecutive residues with same number"):
        topology = md.load(get_fn("two_residues_same_resnum.gro")).topology
    atoms, bonds = topology.to_dataframe()

    topology2 = md.Topology.from_dataframe(atoms, bonds)
    eq(topology, topology2)


def test_topology_numbers(get_fn):
    topology = md.load(get_fn("1bpi.pdb")).topology
    assert len(list(topology.atoms)) == topology.n_atoms
    assert len(list(topology.residues)) == topology.n_residues
    assert all([topology.atom(i).index == i for i in range(topology.n_atoms)])


def test_topology_unique_elements_bpti(get_fn):
    traj = md.load(get_fn("bpti.pdb"))
    top, bonds = traj.top.to_dataframe()
    atoms = np.unique(["C", "O", "N", "H", "S"])
    eq(atoms, np.unique(top.element.values))


def test_chain(get_fn):
    top = md.load(get_fn("bpti.pdb")).topology
    chain = top.chain(0)
    assert chain.n_residues == len(list(chain.residues))

    atoms = list(chain.atoms)
    assert chain.n_atoms == len(atoms)
    for i in range(chain.n_atoms):
        assert atoms[i] == chain.atom(i)


def test_residue(get_fn):
    top = md.load(get_fn("bpti.pdb")).topology
    residue = top.residue(0)
    assert len(list(residue.atoms)) == residue.n_atoms
    atoms = list(residue.atoms)
    for i in range(residue.n_atoms):
        assert residue.atom(i) == atoms[i]


def test_segment_id(get_fn):
    top = md.load(get_fn("ala_ala_ala.pdb")).topology
    assert next(top.residues).segment_id == "AAL", "Segment id is not being assigned correctly for ala_ala_ala.psf"
    df = top.to_dataframe()[0]
    assert len(df["segmentID"] == "AAL") == len(
        df,
    ), "Segment id is not being assigned correctly to topology data frame ala_ala_ala.psf"


def test_nonconsective_resSeq(get_fn):
    t = md.load(get_fn("nonconsecutive_resSeq.pdb"))
    assert eq(np.array([r.resSeq for r in t.top.residues]), np.array([1, 3, 5]))
    df1 = t.top.to_dataframe()
    df2 = md.Topology.from_dataframe(*df1).to_dataframe()
    assert eq(df1[0], df2[0])

    # round-trip through a PDB load/save loop
    fd, fname = tempfile.mkstemp(suffix=".pdb")
    os.close(fd)
    t.save(fname)
    t2 = md.load(fname)
    assert eq(df1[0], t2.top.to_dataframe()[0])
    os.unlink(fname)


def test_pickle(get_fn):
    # test pickling of topology (bug #391)
    topology_without_bond_order = md.load(get_fn("bpti.pdb")).topology
    topology_with_bond_order = md.load(get_fn("imatinib.mol2")).topology
    for top in [topology_with_bond_order, topology_without_bond_order]:
        loaded_top = pickle.loads(pickle.dumps(top))
        assert loaded_top == top


def test_atoms_by_name(get_fn):
    top = md.load(get_fn("bpti.pdb")).topology

    atoms = list(top.atoms)
    for atom1, atom2 in zip(top.atoms_by_name("CA"), top.chain(0).atoms_by_name("CA")):
        assert atom1 == atom2
        assert atom1 in atoms
        assert atom1.name == "CA"

    assert len(list(top.atoms_by_name("CA"))) == sum(1 for _ in atoms if _.name == "CA")
    assert top.residue(15).atom("CA") == [a for a in top.residue(15).atoms if a.name == "CA"][0]

    with pytest.raises(KeyError):
        top.residue(15).atom("sdfsdf")


def test_select_atom_indices(get_fn):
    top = md.load(get_fn("native.pdb")).topology

    assert eq(top.select_atom_indices("alpha"), np.array([8]))
    assert eq(
        top.select_atom_indices("minimal"),
        np.array([4, 5, 6, 8, 10, 14, 15, 16, 18]),
    )

    with pytest.raises(ValueError):
        top.select_atom_indices("sdfsdf")


@needs_openmm
def test_top_dataframe_openmm_roundtrip(get_fn):
    t = md.load(get_fn("2EQQ.pdb"))
    top, bonds = t.top.to_dataframe()
    t.topology = md.Topology.from_dataframe(top, bonds)
    t.top.to_openmm()


def test_n_bonds(get_fn):
    t = md.load(get_fn("2EQQ.pdb"))
    for atom in t.top.atoms:
        if atom.element.symbol == "H":
            assert atom.n_bonds == 1
        elif atom.element.symbol == "C":
            assert atom.n_bonds in [3, 4]
        elif atom.element.symbol == "O":
            assert atom.n_bonds in [1, 2]


def test_load_unknown_topology(get_fn):
    try:
        md.load(get_fn("frame0.dcd"), top=get_fn("frame0.dcd"))
    except OSError as e:
        # we want to make sure there's a nice error message than includes
        # a list of the supported topology formats.
        assert all(s in str(e) for s in (".pdb", ".psf", ".prmtop"))
    else:
        assert False  # fail


def test_unique_pairs():
    n = 10
    a = np.arange(n)
    b = np.arange(n, n + n)

    eq(
        md.Topology._unique_pairs(a, a).sort(),
        md.Topology._unique_pairs_equal(a).sort(),
    )
    eq(
        md.Topology._unique_pairs(a, b).sort(),
        md.Topology._unique_pairs_mutually_exclusive(a, b).sort(),
    )


def test_select_pairs(get_fn):
    traj = md.load(get_fn("tip3p_300K_1ATM.pdb"))
    select_pairs = traj.top.select_pairs

    assert len(select_pairs(selection1="name O", selection2="name O")) == 258 * (258 - 1) // 2
    assert len(select_pairs(selection1="name H1", selection2="name O")) == 258 * 258

    selections = iter(
        [
            # Equal
            ("(name O) or (name =~ 'H.*')", "(name O) or (name =~ 'H.*')"),
            ("all", "all"),
            # Exclusive
            ("name O", "name H1"),
            ("name H1", "name O"),
            # Overlap
            (range(traj.n_atoms), "name O"),
            ("all", "name O"),
        ],
    )

    for select1, select2 in selections:
        select3, select4 = next(selections)
        assert eq(
            select_pairs(selection1=select1, selection2=select2).sort(),
            select_pairs(selection1=select3, selection2=select4).sort(),
        )


def test_to_fasta(get_fn):
    t = md.load(get_fn("2EQQ.pdb"))
    assert t.topology.to_fasta(0) == "ENFSGGCVAGYMRTPDGRCKPTFYQLIT"


def test_subset(get_fn):
    t1 = md.load(get_fn("2EQQ.pdb")).top
    t2 = t1.subset([1, 2, 3])
    assert t2.n_residues == 1


def test_subset_re_index_residues(get_fn):
    t1 = md.load(get_fn("2EQQ.pdb")).top
    t2 = t1.subset(t1.select("resid 0 2"))
    np.testing.assert_array_equal([0, 1], [rr.index for rr in t2.residues])


def test_molecules(get_fn):
    top = md.load(get_fn("4OH9.pdb")).topology
    molecules = top.find_molecules()
    assert sum(len(mol) for mol in molecules) == top.n_atoms
    assert sum(1 for mol in molecules if len(mol) > 1) == 2  # All but two molecules are water


def test_copy_and_hash(get_fn):
    t = md.load(get_fn("traj.h5"))
    t1 = t.topology
    t2 = t.topology.copy()
    assert t1 == t2

    assert hash(tuple(t1._chains)) == hash(tuple(t2._chains))
    assert hash(tuple(t1._atoms)) == hash(tuple(t2._atoms))
    assert hash(tuple(t1._bonds)) == hash(tuple(t2._bonds))
    assert hash(tuple(t1._residues)) == hash(tuple(t2._residues))

    assert hash(t1) == hash(t2)


@pytest.mark.parametrize("filename", ["traj.h5", "1ply_charge.pdb"])
def test_copy(get_fn, filename):
    t = md.load(get_fn(filename))
    t1 = t.topology
    t2 = t.topology.copy()

    assert t1 == t2

    # hash for chains checks only index, but chains have
    # index and chain number, so this isn't as much of a concern
    assert hash(tuple(t1._chains)) == hash(tuple(t2._chains))

    # bond hash checks atom IDs, type and order
    assert hash(tuple(t1._bonds)) == hash(tuple(t2._bonds))

    # Check that atoms are the same
    # because hash isn't complete
    # see issue #2039.
    # hash should be sufficient after this is resolved
    for x, y in zip(t1._atoms, t2._atoms):
        assert x == y, f"Atoms {x} and {y} do not match."


def test_topology_sliced_residue_indices(get_fn):
    # https://github.com/mdtraj/mdtraj/issues/1585
    full = md.load(get_fn("1bpi.pdb"))
    residues = full.top.select("resid 1 to 10")
    sliced = full.atom_slice(residues)
    idx = [res.index for res in sliced.top.residues][-1]
    assert idx == sliced.top.n_residues - 1
    # Now see if this works
    _ = sliced.topology.residue(idx)


def test_topology_join(get_fn):
    top_1 = md.load(get_fn("2EQQ.pdb")).topology
    top_2 = md.load(get_fn("4OH9.pdb")).topology

    out_topology = top_1.join(top_2)

    eq(out_topology.n_atoms, top_1.n_atoms + top_2.n_atoms)
    eq(out_topology.n_residues, top_1.n_residues + top_2.n_residues)
    eq(top_1.atom(0).residue.name, out_topology.atom(0).residue.name)
    eq(top_2.atom(-1).residue.name, out_topology.atom(-1).residue.name)
    eq(top_1.atom(0).element, out_topology.atom(0).element)
    eq(top_2.atom(-1).element, out_topology.atom(-1).element)


def test_topology_join_keep_resSeq(get_fn):
    top_1 = md.load(get_fn("2EQQ.pdb")).topology
    top_2 = md.load(get_fn("4OH9.pdb")).topology

    out_topology_keepId_True = top_1.join(top_2, keep_resSeq=True)
    out_topology_keepId_False = top_1.join(top_2, keep_resSeq=False)

    out_resSeq_keepId_True = [residue.resSeq for residue in out_topology_keepId_True.residues]
    out_resSeq_keepId_False = [residue.resSeq for residue in out_topology_keepId_False.residues]

    expected_resSeq_keepId_True = [residue.resSeq for residue in top_1.residues] + [
        residue.resSeq for residue in top_2.residues
    ]
    expected_resSeq_keepId_False = list(range(1, len(expected_resSeq_keepId_True) + 1))

    eq(out_resSeq_keepId_True, expected_resSeq_keepId_True)
    eq(out_resSeq_keepId_False, expected_resSeq_keepId_False)


def test_topology_join_formal_charge():
    top1 = md.Topology()
    c1 = top1.add_chain()
    r1 = top1.add_residue("LIG", c1, 0)
    top1.add_atom("C1", md.element.Element.getBySymbol("C"), r1, formal_charge=1)

    top2 = md.Topology()
    c2 = top2.add_chain()
    r2 = top2.add_residue("LIG", c2, 0)
    top2.add_atom("C2", md.element.Element.getBySymbol("C"), r2, formal_charge=-1)

    joined = top1.join(top2)

    charges = [atom.formal_charge for atom in joined.atoms]
    assert charges == [1, -1], f"Expected [1, -1] formal charges, got {charges}"


def test_delete_atom_by_index(get_fn):
    """
    REMARK *
    REMARK   DATE:     8/ 5/ 9     14:44:19      CREATED BY USER: mjw
    ATOM      1  N   ALA     1       0.024  -0.103  -0.101  1.00  0.00      AAL
    ATOM      2  HT1 ALA     1       0.027  -1.132  -0.239  1.00  0.00      AAL
    ATOM      3  HT2 ALA     1      -0.805   0.163   0.471  1.00  0.00      AAL
    ATOM      4  HT3 ALA     1      -0.059   0.384  -1.019  1.00  0.00      AAL
    ATOM      5  CA  ALA     1       1.247   0.375   0.636  1.00  0.00      AAL
    ...
    """
    index = 0
    top = md.load(get_fn("ala_ala_ala.pdb")).topology
    n_bonds_original = top.n_bonds
    n_atoms_original = top.n_atoms
    atom_to_remove = top._atoms[index]
    top.delete_atom_by_index(index)
    assert top.n_atoms == n_atoms_original - 1
    assert top.n_bonds == n_bonds_original - 4
    assert all([atom_to_remove not in bond for bond in top.bonds])
    assert eq(list(range(top.n_atoms)), [atom.index for atom in top.atoms])



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2022 Stanford University and the Authors
#
# Authors: Peter Eastman, Robert McGibbon
# Contributors: Carlos Hernandez, Jason Swails
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
# Portions copyright (c) 2012 Stanford University and the Authors.
#
# Portions of this code originate from the OpenMM molecular simulation
# toolkit. Those portions are Copyright 2008-2012 Stanford University
# and Peter Eastman, and distributed under the following license:
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS, CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
# USE OR OTHER DEALINGS IN THE SOFTWARE.
##############################################################################

import os
import warnings

import numpy as np

from mdtraj.formats.registry import FormatRegistry
from mdtraj.utils import cast_indices, in_units_of, open_maybe_zipped

__all__ = ["load_pdbx", "PDBxTrajectoryFile"]


##############################################################################
# Code
##############################################################################


@FormatRegistry.register_loader(".cif")
@FormatRegistry.register_loader(".cif.gz")
@FormatRegistry.register_loader(".mmcif")
@FormatRegistry.register_loader(".mmcif.gz")
@FormatRegistry.register_loader(".pdbx")
@FormatRegistry.register_loader(".pdbx.gz")
def load_pdbx(
    filename,
    stride=None,
    atom_indices=None,
    frame=None,
    no_boxchk=False,
    top=None,
):
    """Load a PDBx/mmCIF file from disk.

    Parameters
    ----------
    filename : path-like
        Path to the PDBx/mmCIF file on disk
    stride : int, default=None
        Only read every stride-th model from the file
    atom_indices : array_like, default=None
        If not None, then read only a subset of the atoms coordinates from the
        file. These indices are zero-based (not 1 based, as used by the PDBx/mmCIF
        format). So if you want to load only the first atom in the file, you
        would supply ``atom_indices = np.array([0])``.
    frame : int, default=None
        Use this option to load only a single frame from a trajectory on disk.
        If frame is None, the default, the entire trajectory will be loaded.
        If supplied, ``stride`` will be ignored.
    no_boxchk : bool, default=False
        By default, a heuristic check based on the particle density will be
        performed to determine if the unit cell dimensions are absurd. If the
        particle density is >1000 atoms per nm^3, the unit cell will be
        discarded. This is done because all PDBx/mmCIF files from RCSB contain a ``cell``
        record, even if there are no periodic boundaries, and dummy values are
        filled in instead. This check will filter out those false unit cells and
        avoid potential errors in geometry calculations. Set this variable to
        ``True`` in order to skip this heuristic check.
    top : mdtraj.core.Topology, default=None
        if you give a topology as input the topology won't be parsed from the file

    Returns
    -------
    trajectory : md.Trajectory
        The resulting trajectory, as an md.Trajectory object.

    Examples
    --------
    >>> import mdtraj as md
    >>> pdbx = md.load_pdbx('2EQQ.pdbx')
    >>> print(pdbx)
    <mdtraj.Trajectory with 20 frames, 423 atoms at 0x110740a90>

    See Also
    --------
    mdtraj.PDBxTrajectoryFile : Low level interface to PDBx/mmCIF files
    """
    from mdtraj import Trajectory

    if not isinstance(filename, (str, os.PathLike)):
        raise TypeError(
            "filename must be of type string or path-like for load_pdb. You supplied %s" % type(filename),
        )

    atom_indices = cast_indices(atom_indices)

    with PDBxTrajectoryFile(filename, top=top) as f:
        atom_slice = slice(None) if atom_indices is None else atom_indices
        if frame is not None:
            coords = f.positions[[frame], atom_slice, :]
        else:
            coords = f.positions[::stride, atom_slice, :]
        assert coords.ndim == 3, "internal shape error"
        n_frames = len(coords)

        topology = f.topology
        if atom_indices is not None:
            # The input topology shouldn't be modified because
            # subset makes a copy inside the function
            topology = topology.subset(atom_indices)

        if f.unitcell_angles is not None and f.unitcell_lengths is not None:
            unitcell_lengths = np.array([f.unitcell_lengths] * n_frames)
            unitcell_angles = np.array([f.unitcell_angles] * n_frames)
        else:
            unitcell_lengths = None
            unitcell_angles = None

        in_units_of(coords, f.distance_unit, Trajectory._distance_unit, inplace=True)
        in_units_of(
            unitcell_lengths,
            f.distance_unit,
            Trajectory._distance_unit,
            inplace=True,
        )

    time = np.arange(len(coords))
    if frame is not None:
        time *= frame
    elif stride is not None:
        time *= stride

    traj = Trajectory(
        xyz=coords,
        time=time,
        topology=topology,
        unitcell_lengths=unitcell_lengths,
        unitcell_angles=unitcell_angles,
    )

    if not no_boxchk and traj.unitcell_lengths is not None:
        # Some PDBx/mmCIF files do not *really* have a unit cell, but still
        # have a cell record with a dummy definition. These boxes are usually
        # tiny (e.g., 1 A^3), so check that the particle density in the unit
        # cell is not absurdly high. Standard water density is ~55 M, which
        # yields a particle density ~100 atoms per cubic nm. It should be safe
        # to say that no particle density should exceed 10x that.
        particle_density = traj.top.n_atoms / traj.unitcell_volumes[0]
        if particle_density > 1000:
            warnings.warn(
                "Unlikely unit cell vectors detected in PDB file likely "
                "resulting from a dummy CRYST1 record. Discarding unit "
                "cell vectors.",
                category=UserWarning,
            )
            traj._unitcell_lengths = traj._unitcell_angles = None

    return traj


@FormatRegistry.register_fileobject(".cif")
@FormatRegistry.register_fileobject(".cif.gz")
@FormatRegistry.register_fileobject(".mmcif")
@FormatRegistry.register_fileobject(".mmcif.gz")
@FormatRegistry.register_fileobject(".pdbx")
@FormatRegistry.register_fileobject(".pdbx.gz")
class PDBxTrajectoryFile:
    """Interface for reading and writing PDBx/mmCIF files

    Parameters
    ----------
    filename : path-like
        The filename to open. A path to a file on disk.
    mode : {'r', 'w'}
        The mode in which to open the file, either 'r' for read or 'w' for write.
    force_overwrite : bool
        If opened in write mode, and a file by the name of `filename` already
        exists on disk, should we overwrite it?
    top : mdtraj.core.Topology, default=None
        if you give a topology as input the topology won't be parsed from the file

    Attributes
    ----------
    positions : np.ndarray, shape=(n_frames, n_atoms, 3)
    topology : mdtraj.Topology
    closed : bool

    See Also
    --------
    mdtraj.load_pdbx : High-level wrapper that returns a ``md.Trajectory``
    """

    distance_unit = "nanometers"

    def __init__(self, filename, mode="r", force_overwrite=True, top=None):
        self._open = False
        self._mode = mode
        from .pdbxfile import PDBxFile

        if mode == "r":
            self._open = True
            pdbx = PDBxFile(filename)
            if top is None:
                self._topology = pdbx.topology
            else:
                self._topology = top
            positions = [pdbx.getPositions(asNumpy=True, frame=i) for i in range(pdbx.getNumFrames())]
            self._positions = np.array(positions)
            if pdbx._unitcell_angles is not None and pdbx._unitcell_lengths is not None:
                l1, l2, l3 = pdbx._unitcell_lengths
                alpha, beta, gamma = pdbx._unitcell_angles
                self._unitcell_lengths = (l1, l2, l3)
                self._unitcell_angles = (alpha, beta, gamma)
            else:
                self._unitcell_lengths = None
                self._unitcell_angles = None
        elif mode == "w":
            self._open = True
            self._next_model = 0
            self._file = open_maybe_zipped(filename, "w", force_overwrite)
        else:
            raise ValueError("invalid mode: %s" % mode)

    def write(
        self,
        positions,
        topology,
        unitcell_lengths=None,
        unitcell_angles=None,
        bfactors=None,
        keepIds=False,
    ):
        """Write one frame of a molecular dynamics trajectory to disk in PDBx/mmCIF format.

        Parameters
        ----------
        positions : array_like
            The list of atomic positions to write in nanometers.
        topology : mdtraj.Topology
            The Topology defining the model to write.
        unitcell_lengths : {tuple, None}
            Lengths of the three unit cell vectors, or None for a non-periodic system
        unitcell_angles : {tuple, None}
            Angles between the three unit cell vectors, or None for a non-periodic system
        bfactors : array_like, default=None, shape=(n_atoms,)
            Save bfactors with cif file. Should contain a single number for each atom in the topology
        keepIds : bool=False
            If True, keep the residue and chain IDs specified in the Topology
            rather than generating new ones.  Warning: It is up to the caller to
            make sure these are valid IDs that satisfy the requirements of the
            PDBx/mmCIF format.  Otherwise, the output file will be invalid.
        """
        if not self._mode == "w":
            raise ValueError("file not opened for writing")
        from .pdbxfile import PDBxFile

        if self._next_model == 0:
            PDBxFile.writeHeader(
                topology,
                unitcell_lengths,
                unitcell_angles,
                self._file,
                keepIds=keepIds,
            )
            self._next_model = 1
        if len(positions.shape) == 3:
            positions = positions[0]
        PDBxFile.writeModel(
            topology,
            positions,  # nanometers
            self._file,
            self._next_model,
            bfactors=bfactors,
            keepIds=keepIds,
        )
        self._next_model += 1

    @property
    def positions(self):
        """The cartesian coordinates of all of the atoms in each frame. Available when a file is opened in mode='r'"""
        return self._positions

    @property
    def topology(self):
        """The topology from this PDBx/mmCIF file. Available when a file is opened in mode='r'"""
        return self._topology

    @property
    def unitcell_lengths(self):
        """The unitcell lengths (3-tuple) in this PDBx/mmCIF file. May be None"""
        return self._unitcell_lengths

    @property
    def unitcell_angles(self):
        """The unitcell angles (3-tuple) in this PDBx/mmCIF file. May be None"""
        return self._unitcell_angles

    @property
    def closed(self):
        """Whether the file is closed"""
        return not self._open

    def close(self):
        """Close the PDBx/mmCIF file"""
        if self._mode == "w" and self._open:
            self._file.close()
        self._open = False

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

    def __len__(self):
        "Number of frames in the file"
        if str(self._mode) != "r":
            raise NotImplementedError('len() only available in mode="r" currently')
        return len(self._positions)



import bz2
import gzip

from mdtraj.testing import eq
from mdtraj.utils import open_maybe_zipped


def test_read_gz(tmpdir):
    fn = f"{tmpdir}/read.gz"
    with gzip.GzipFile(fn, "w") as f:
        f.write(b"COOKIE")
    eq(open_maybe_zipped(fn, "r").read(), "COOKIE")


def test_write_gz(tmpdir):
    fn = f"{tmpdir}/write.gz"
    with open_maybe_zipped(fn, "w") as f:
        f.write("COOKIE")
    with gzip.GzipFile(fn, "r") as f:
        eq(f.read().decode("utf-8"), "COOKIE")


def test_read_bz2(tmpdir):
    fn = f"{tmpdir}/read.bz2"
    with bz2.BZ2File(fn, "w") as f:
        f.write(b"COOKIE")
    eq(open_maybe_zipped(fn, "r").read(), "COOKIE")


def test_write_bz2(tmpdir):
    fn = f"{tmpdir}/write.bz2"
    with open_maybe_zipped(fn, "w") as f:
        f.write("COOKIE")
    with bz2.BZ2File(fn, "r") as f:
        eq(f.read().decode("utf-8"), "COOKIE")



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2017 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import itertools

import numpy as np
import pytest

import mdtraj as md
from mdtraj.geometry.distance import (
    _displacement,
    _displacement_mic,
    compute_displacements,
    compute_distances,
    compute_distances_core,
    compute_distances_t,
    find_closest_contact,
)
from mdtraj.testing import assert_allclose, eq


class Test_Distance:
    def test_generator(self, gen_random_ptraj):
        self.pairs2 = itertools.combinations(range(self.N_ATOMS), 2)
        a = compute_distances(self.ptraj, self.pairs)
        b = compute_distances(self.ptraj, self.pairs2)
        eq(a, b)

    def test_0(self, gen_random_ptraj):
        a = compute_distances(self.ptraj, self.pairs, periodic=False, opt=True)
        b = compute_distances(self.ptraj, self.pairs, periodic=False, opt=False)
        eq(a, b)

    def test_compute_distances_core_nonperiodic(self, gen_random_ptraj):
        a = compute_distances(self.ptraj, self.pairs, periodic=False, opt=True)
        b = compute_distances_core(
            self.ptraj.xyz,
            self.pairs,
            unitcell_vectors=self.ptraj.unitcell_vectors,
            periodic=False,
            opt=True,
        )
        eq(a, b)

        a = compute_distances(self.ptraj, self.pairs, periodic=False, opt=False)
        b = compute_distances_core(
            self.ptraj.xyz,
            self.pairs,
            unitcell_vectors=self.ptraj.unitcell_vectors,
            periodic=False,
            opt=False,
        )
        eq(a, b)

    def test_compute_distances_core_periodic(self, gen_random_ptraj):
        # opt
        a = compute_distances(self.ptraj, self.pairs, periodic=True, opt=True)
        b = compute_distances_core(
            self.ptraj.xyz,
            self.pairs,
            unitcell_vectors=self.ptraj.unitcell_vectors,
            periodic=True,
            opt=True,
        )
        eq(a, b, decimal=3)

        # no-opt
        a = compute_distances(self.ptraj, self.pairs, periodic=True, opt=False)
        b = compute_distances_core(
            self.ptraj.xyz,
            self.pairs,
            unitcell_vectors=self.ptraj.unitcell_vectors,
            periodic=True,
            opt=False,
        )
        eq(a, b, decimal=3)

    def test_1(self, gen_random_ptraj):
        a = compute_displacements(self.ptraj, self.pairs, periodic=False, opt=True)
        b = compute_displacements(self.ptraj, self.pairs, periodic=False, opt=False)
        eq(a, b)

    def test_2(self, gen_random_ptraj):
        a = compute_distances(self.ptraj, self.pairs, periodic=False, opt=False)
        b = compute_displacements(self.ptraj, self.pairs, periodic=False, opt=False)
        eq(a, np.sqrt(np.sum(np.square(b), axis=2)))

    def test_3(self, gen_random_ptraj):
        a = compute_distances(self.ptraj, self.pairs, periodic=False, opt=True)
        b = compute_displacements(self.ptraj, self.pairs, periodic=False, opt=True)
        eq(a, np.sqrt(np.sum(np.square(b), axis=2)))

    def test_0p(self, gen_random_ptraj):
        a = compute_distances(self.ptraj, self.pairs, periodic=True, opt=True)
        b = compute_distances(self.ptraj, self.pairs, periodic=True, opt=False)
        eq(a, b, decimal=3)

    def test_1p(self, gen_random_ptraj):
        a = compute_displacements(self.ptraj, self.pairs, periodic=True, opt=True)
        b = compute_displacements(self.ptraj, self.pairs, periodic=True, opt=False)
        eq(a, b, decimal=3)

    def test_2p(self, gen_random_ptraj):
        a = compute_distances(self.ptraj, self.pairs, periodic=True, opt=False)
        b = compute_displacements(self.ptraj, self.pairs, periodic=True, opt=False)
        assert a.shape == (len(self.ptraj), len(self.pairs))
        assert b.shape == (len(self.ptraj), len(self.pairs), 3), str(b.shape)
        b = np.sqrt(np.sum(np.square(b), axis=2))
        eq(a, b, decimal=5)

    def test_3p(self, gen_random_ptraj):
        a = compute_distances(self.ptraj, self.pairs, periodic=True, opt=True)
        b = compute_displacements(self.ptraj, self.pairs, periodic=True, opt=True)
        eq(a, np.sqrt(np.sum(np.square(b), axis=2)))

    def test_4(self, gen_random_ptraj):
        # using a really big box, we should get the same results with and without
        # pbcs
        box = np.array([[100, 0, 0], [0, 200, 0], [0, 0, 300]])
        box = np.zeros((self.N_FRAMES, 3, 3)) + box  # broadcast it out
        a = _displacement_mic(self.xyz, self.pairs, box, False)
        b = _displacement(self.xyz, self.pairs)
        eq(a, b, decimal=3)

    def test_5(self, gen_random_ptraj):
        # simple wrap around along the z axis.
        xyz = np.array([[[0.0, 0.0, 0.0], [0.0, 0.0, 2.2]]])
        box = np.eye(3, 3).reshape(1, 3, 3)
        result = _displacement_mic(xyz, np.array([[0, 1]]), box, True)
        eq(result, np.array([[[0, 0, 0.2]]]))

    def test_6(self, get_fn):
        ext_ref = (
            np.array(
                [
                    17.4835,
                    22.2418,
                    24.2910,
                    22.5505,
                    12.8686,
                    22.1090,
                    7.4472,
                    22.4253,
                    19.8283,
                    20.6935,
                ],
            )
            / 10
        )
        traj = md.load(get_fn("test_good.nc"), top=get_fn("test.parm7"))
        self._run_amber_traj(traj, ext_ref)

    def test_7(self, get_fn):
        ext_ref = (
            np.array(
                [
                    30.9184,
                    23.9040,
                    25.3869,
                    28.0060,
                    25.9704,
                    24.6836,
                    23.0508,
                    27.1983,
                    24.4954,
                    26.7448,
                ],
            )
            / 10
        )
        traj = md.load(get_fn("test_bad.nc"), top=get_fn("test.parm7"))
        self._run_amber_traj(traj, ext_ref)

    def _run_amber_traj(self, traj, ext_ref):
        # Test triclinic case where simple approach in Tuckerman text does not
        # always work
        distopt = md.compute_distances(traj, [[0, 9999]], opt=True)
        distslw = md.compute_distances(traj, [[0, 9999]], opt=False)
        dispopt = md.compute_displacements(traj, [[0, 9999]], opt=True)
        dispslw = md.compute_displacements(traj, [[0, 9999]], opt=False)

        eq(distopt, distslw, decimal=5)
        eq(dispopt, dispslw, decimal=5)

        assert_allclose(distopt.flatten(), ext_ref, atol=2e-5)

        # Make sure distances from displacements are the same
        eq(np.sqrt((dispopt.squeeze() ** 2).sum(axis=1)), distopt.squeeze())
        eq(np.sqrt((dispslw.squeeze() ** 2).sum(axis=1)), distslw.squeeze())
        eq(dispopt, dispslw, decimal=5)

    def test_closest_contact(self, gen_random_ptraj):
        box_size = np.array([3.0, 4.0, 5.0])
        traj = md.Trajectory(xyz=self.xyz * box_size, topology=None)
        self._verify_closest_contact(gen_random_ptraj, traj)
        traj.unitcell_lengths = np.array([box_size for i in range(self.N_FRAMES)])
        traj.unitcell_angles = np.array([[90.0, 90.0, 90.0] for i in range(self.N_FRAMES)])
        self._verify_closest_contact(gen_random_ptraj, traj=traj)
        traj.unitcell_angles = np.array([[80.0, 90.0, 100.0] for i in range(self.N_FRAMES)])
        self._verify_closest_contact(gen_random_ptraj, traj=traj)

    def _verify_closest_contact(self, gen_random_ptraj, traj):
        group1 = np.array([i for i in range(self.N_ATOMS // 2)], dtype=int)
        group2 = np.array([i for i in range(self.N_ATOMS // 2, self.N_ATOMS)], dtype=int)
        contact = find_closest_contact(traj, group1, group2)
        pairs = np.array([(i, j) for i in group1 for j in group2], dtype=int)
        dists = md.compute_distances(traj, pairs, True)[0]
        _ = md.compute_distances(traj, pairs, False)[0]
        nearest = np.argmin(dists)
        eq(float(dists[nearest]), contact[2], decimal=5)
        assert (pairs[nearest, 0] == contact[0] and pairs[nearest, 1] == contact[1]) or (
            pairs[nearest, 0] == contact[1] and pairs[nearest, 1] == contact[0]
        )

    def test_distance_nan(self):
        xyz = np.array([[1, 1, 1], [2, 1, 1], [np.nan, np.nan, np.nan]]).reshape(1, 3, 3)
        dists = md.compute_distances(md.Trajectory(xyz=xyz, topology=None), [[0, 1]])
        assert np.isfinite(dists).all()

    def test_closest_contact_nan_pos(self, gen_random_ptraj):
        box_size = np.array([3.0, 4.0, 5.0])
        xyz = self.rng.standard_normal((2, 20, 3), dtype=np.float32)
        xyz *= box_size
        # Set the last frame to nan
        xyz[-1] = np.nan
        # Slice of the last frame, so nans should not cause troubles.
        xyz = xyz[:-1]
        traj = md.Trajectory(xyz=xyz, topology=None)
        self._verify_closest_contact(gen_random_ptraj, traj)

    def test_distance_t_inputs(self, gen_random_ptraj):
        incorrect_pairs = np.array((0, self.ptraj.n_atoms + 1))
        with pytest.raises(ValueError, match="atom_pairs"):
            compute_distances_t(self.ptraj, incorrect_pairs, self.times)

        incorrect_times = np.array((0, self.ptraj.n_frames + 1))
        with pytest.raises(ValueError, match="time_pairs"):
            compute_distances_t(self.ptraj, self.pairs, incorrect_times)

    def test_distances_t(self, gen_random_ptraj):
        a = compute_distances_t(self.ptraj, self.pairs, self.times, periodic=True, opt=True)
        b = compute_distances_t(self.ptraj, self.pairs, self.times, periodic=True, opt=False)
        # There is a precision difference between cython and python code for ``compute_distances_t``
        # on linux-aarch64 for certain seeds (e.g. seed=3220). Pinning for a lower precision check.
        eq(a, b, decimal=4)

        c = compute_distances_t(self.ptraj, self.pairs, self.times, periodic=False, opt=True)
        d = compute_distances_t(self.ptraj, self.pairs, self.times, periodic=False, opt=False)
        eq(c, d)

    def test_distances_t_at_0(self, gen_random_ptraj):
        self.times = np.array([[0, 0]], dtype=np.int32)
        a = compute_distances_t(self.ptraj, self.pairs, self.times, periodic=True, opt=True)
        b = compute_distances_t(self.ptraj, self.pairs, self.times, periodic=True, opt=False)
        c = compute_distances(self.ptraj[:1], self.pairs, periodic=True, opt=True)
        d = compute_distances(self.ptraj[:1], self.pairs, periodic=True, opt=False)
        eq(a, c)
        eq(b, d)

    def _run_amber_traj_t(self, traj, ext_ref):
        # Test triclinic case where simple approach in Tuckerman text does not
        # always work
        _ = compute_distances_t(
            traj,
            atom_pairs=[[0, 9999]],
            time_pairs=[[0, 2]],
            opt=True,
        )
        _ = compute_distances_t(
            traj,
            atom_pairs=[[0, 9999]],
            time_pairs=[[0, 2]],
            opt=False,
        )

    def test_amber_t(self, get_fn):
        ext_ref = (
            np.array(
                [
                    17.4835,
                    22.2418,
                    24.2910,
                    22.5505,
                    12.8686,
                    22.1090,
                    7.4472,
                    22.4253,
                    19.8283,
                    20.6935,
                ],
            )
            / 10
        )
        traj = md.load(get_fn("test_good.nc"), top=get_fn("test.parm7"))
        self._run_amber_traj_t(traj, ext_ref)



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


import importlib

##############################################################################
# Imports
##############################################################################
import pkgutil
import warnings
from inspect import (
    getargs,
    getdoc,
    getmembers,
    getmodule,
    isbuiltin,
    isclass,
    isfunction,
    ismethod,
    ismodule,
)

from mdtraj.testing.docscrape import NumpyDocString

__all__ = ["docstring_verifiers", "import_all_modules"]

##############################################################################
# functions
##############################################################################


def docstring_verifiers(module, error_on_none=False):
    """Yield an iterable of tests that verify the docstrings (for
    adherance to the NumpyDoc format) of all functions/classes defined
    in a module.

    Parameters
    ----------
    module : module
        The module to test
    error_on_none : bool, default=False
        Throw an error if no docstring is defined
    """

    # These are the types that we want to check
    # currently, the docstring on classes are not being checked, since
    # isclass() is not in the list
    acceptors = [isfunction, ismethod, isbuiltin]

    def accept(f):
        return any([acc(f) for acc in acceptors])

    functions = [f for f in walk(module) if accept(f)]

    def format(f):
        """
        Format a method/function/class as a string

        Parameters
        ----------
        f : function, method, class

        Returns
        -------
        repr : string
            A string represntation
        """
        if ismethod(f):
            return ".".join([getmodule(f).__name__, f.im_class.__name__, f.__name__])
        if isfunction(f) or isbuiltin(f):
            return ".".join([getmodule(f).__name__, f.__name__])
        if isclass(f):
            return f.__name__
        return "Error"

    def check_docstring(f):
        """
        Ensure the docstring of `f` is in accordance with the numpy standard

        Currently, only the Parameters section of the docstring is checked.

        Parameters
        ----------
        f : function, method, class

        Returns
        -------
        repr : string
            A string represntation
        """

        doc = getdoc(f)
        if doc is None:
            if error_on_none:
                raise ValueError("no docstring for %s" % format(f))
        else:
            with warnings.catch_warnings():
                warnings.simplefilter("error")
                parsed = NumpyDocString(doc)

            param_names = {e[0] for e in parsed["Parameters"]}

            if isbuiltin(f):
                # You can't get the arglist from a builtin, which
                # is how cython functions turn up

                # but you can, hackily, get the number of arguments it wants
                # by parseing the error hen you supply too many
                import re

                try:
                    f(*list(range(100)))
                except TypeError as e:
                    m = re.search(r"takes at most (\d+) positional arguments", str(e))
                    if not m:
                        return
                    n_args = int(m.group(1))

                if len(param_names) != n_args:
                    raise ValueError(
                        "In %s, number of arguments, %d, doesn't "
                        " match the length of the Parameters in the "
                        "docstring, %d" % (format(f), n_args, len(param_names)),
                    )
                return

            args = set(getargs(f.__code__).args)

            if "self" in args:
                args.remove("self")
            if "cls" in args:
                args.remove("cls")

            if args != param_names:
                raise ValueError(
                    f"In {format(f)}, arguments {list(args)} don't match Parameters list {list(param_names)}",
                )

    for f in functions:

        def qq():
            return check_docstring(f)

        qq.description = f"NumpyDoc: {module.__name__}.{f.__name__}"
        qq.fname = f.__name__
        yield qq


def ispackage(obj):
    """
    Check if obj is a package. Simply look for whether its a module whose
    filename is __init__.py(c)

    Parameters
    ----------
    obj : module
    """
    if ismodule(obj):
        return obj.__file__.endswith("__init__.pyc") or obj.__file__.endswith(
            "__init__.py",
        )
    return False


def walk(module):
    """
    Get all of the functions, classes and their methods defined within
    a python module

    Parameters
    ----------
    module : module

    Returns
    -------
    funcs : list
        List of functions, classes and methods
    """
    assert ismodule(module)
    if ispackage(module):
        raise ValueError("Sorry, you need to supply me a module, not a package")

    def is_valid(obj):
        if getmodule(obj) == module:
            # cython specific stuff
            if module.__file__.endswith(".so"):
                if isbuiltin(obj):
                    return not obj.__name__.startswith("_")

            if ismethod(obj) or isfunction(obj):
                return not obj.__name__.startswith("_")
            if isclass(obj):
                return True
        return False

    instack = [v for k, v in getmembers(module) if is_valid(v)]
    outstack = []

    while True:
        try:
            item = instack.pop()
        except IndexError:
            break
        outstack.append(item)

        if isclass(item):
            instack.extend([v for k, v in getmembers(item) if is_valid(v)])

    return outstack


def import_all_modules(pkg):
    result = []
    for _, modname, ispkg in pkgutil.iter_modules(pkg.__path__):
        c = f"{pkg.__name__}.{modname}"
        if modname.startswith("test_"):
            continue
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=DeprecationWarning)
                mod = importlib.import_module(c)
            if ispkg:
                result.extend(import_all_modules(mod))
            else:
                result.append(mod)
        except ImportError as e:
            print("e", e)
            continue

    return result



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2017 Stanford University and the Authors
#
# Authors: Robert McGibbon
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################

import mdtraj as md
from mdtraj.geometry.pi_stacking import pi_stacking


def test_pi_stacking(get_fn):
    """Test that pi-stacking is being detected correctly."""
    t: md.Trajectory = md.load(get_fn("4BFQ.pdb"))
    lig_aromatic_atms_to_test = [
        ("C16", "C17", "C18", "C19", "C20", "C21"),  # pi-stacking
        ("N4", "C9", "C15", "C16", "C21", "C8"),  # pi-stacking
        ("N5", "C10", "C11", "C12", "C13", "C14"),  # NOT pi-stacking
    ]
    lig_grps = []
    for lig_grp in lig_aromatic_atms_to_test:
        lig_grps.append(
            tuple(
                int(idx)
                for idx in t.top.select(
                    f"chainid 5 and resSeq 301 and (name {lig_grp[0]} or name {' or name '.join(lig_grp[1:])})",
                )
            ),
        )
    protein_grps = []
    # TYR186
    protein_stacking_atms = [("CE2", "CD2", "CG", "CD1", "CE1", "CZ")]
    for protein_grp in protein_stacking_atms:
        protein_grps.append(
            tuple(
                int(idx)
                for idx in t.top.select(
                    f"chainid 0 and resSeq 186 and (name {protein_grp[0]} or name {' or name '.join(protein_grp[1:])})",  # noqa
                )
            ),
        )
    stacking_interactions = pi_stacking(t, lig_grps, protein_grps)
    # Since it returns per-frame, just get the 'first'
    stacking_interactions = stacking_interactions[0]
    assert len(stacking_interactions) == 2
    assert (lig_grps[0], protein_grps[0]) in stacking_interactions
    assert (lig_grps[1], protein_grps[0]) in stacking_interactions
    assert (lig_grps[2], protein_grps[0]) not in stacking_interactions


def test_t_stacking_more_relaxed_radius(get_fn):
    """Test that t-stacking version of pi-stacking is being detected correctly."""
    t: md.Trajectory = md.load(get_fn("4BFQ.pdb"))
    lig_aromatic_atms_to_test = [
        ("C2", "C3", "C4", "N1", "C6", "N6"),  # t-stacking
        ("N5", "C10", "C11", "C12", "C13", "C14"),  # NOT pi-stacking
    ]
    lig_grps = []
    for lig_grp in lig_aromatic_atms_to_test:
        lig_grps.append(
            tuple(
                int(idx)
                for idx in t.top.select(
                    f"chainid 5 and resSeq 302 and (name {lig_grp[0]} or name {' or name '.join(lig_grp[1:])})",
                )
            ),
        )
    protein_grps = []
    # TYR91
    protein_stacking_atms = [("CE2", "CD2", "CG", "CD1", "CE1", "CZ")]
    for protein_grp in protein_stacking_atms:
        protein_grps.append(
            tuple(
                int(idx)
                for idx in t.top.select(
                    f"chainid 0 and resSeq 91 and (name {protein_grp[0]} or name {' or name '.join(protein_grp[1:])})",
                )
            ),
        )
    # More permissive edge intersection radius for t-stack
    stacking_interactions = pi_stacking(t, lig_grps, protein_grps, edge_intersection_radius=0.21)
    stacking_interactions = stacking_interactions[0]
    assert len(stacking_interactions) == 1
    assert (lig_grps[0], protein_grps[0]) in stacking_interactions
    assert (lig_grps[1], protein_grps[0]) not in stacking_interactions


def test_t_stacking_default_settings(get_fn):
    """Test that t-stacking version of pi-stacking is being detected correctly."""

    t: md.Trajectory = md.load(get_fn("6A22.pdb"))
    lig_aromatic_atms_to_test = [
        ("S10", "C9", "C8", "N7", "C6"),  # t-stacking
        ("C14", "C15", "C16", "C17", "C18", "C19"),  # NOT pi-stacking
    ]
    lig_grps = []
    for lig_grp in lig_aromatic_atms_to_test:
        lig_grps.append(
            tuple(
                int(idx)
                for idx in t.top.select(
                    f"chainid 9 and resSeq 9000 and (name {lig_grp[0]} or name {' or name '.join(lig_grp[1:])})",
                )
            ),
        )
    protein_grps = []
    # PHR118/378
    protein_stacking_atms = [("CE2", "CD2", "CG", "CD1", "CE1", "CZ")]
    for protein_grp in protein_stacking_atms:
        protein_grps.append(
            tuple(
                int(idx)
                for idx in t.top.select(
                    f"chainid 2 and resSeq 378 and (name {protein_grp[0]} or name {' or name '.join(protein_grp[1:])})",  # noqa
                )
            ),
        )
    stacking_interactions = pi_stacking(t, lig_grps, protein_grps)
    stacking_interactions = stacking_interactions[0]
    assert len(stacking_interactions) == 1
    assert (lig_grps[0], protein_grps[0]) in stacking_interactions
    assert (lig_grps[1], protein_grps[0]) not in stacking_interactions



##############################################################################
# MDTraj: A Python Library for Loading, Saving, and Manipulating
#         Molecular Dynamics Trajectories.
# Copyright 2012-2013 Stanford University and the Authors
#
# Authors: Ryan DiRisio
# Contributors:
#
# MDTraj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 2.1
# of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with MDTraj. If not, see <http://www.gnu.org/licenses/>.
##############################################################################


from collections import defaultdict

import numpy as np

import mdtraj as md
from mdtraj.geometry import compute_distances_core

__all__ = ["pi_stacking"]


def compute_centroid(trajectory, atoms):
    """
    Compute the centroid (center of geometry) of the ring atoms in the trajectory.

    NOTE: Assumes the atoms are not across periodic boundaries.

    Parameters
    ----------
    trajectory : md.Trajectory
        The trajectory to analyze, uses just the xyz coordinates.
    atoms : array_like
        The atom indices involved in the ring.

    Returns
    -------
    centroid : ndarray, shape=[n_frames, 3], dtype=float
        The mean (centroid) of the ring atoms coordinates. MDTraj stores the coordinates
        in nm.
    """
    trj_subset = trajectory.xyz[:, atoms]
    centroid = np.mean(trj_subset, axis=1)
    return centroid


def compute_ring_normal(trajectory, atoms, centroid):
    """
    Calculate the normal vector to the plane defined by the ring atoms and the centroid.

    Parameters
    ----------
    trajectory : md.Trajectory
        The trajectory to analyze, uses just the xyz coordinates.
    atoms : array_like
        The atom indices involved in the ring. Must be at least of length 2.
    centroid : ndarray
        The centroid of the ring atoms.

    Returns
    -------
    norm_vec : np.array, shape=[n_frames, 2, 3], dtype=float

        The normal vector to the plane defined by the ring atoms and the centroid.
        The normal vector is normalized, and the directionality is arbitrarily assigned
        by the order of the atoms list.
    """
    cross = np.cross(
        trajectory.xyz[:, atoms[0]] - centroid,
        trajectory.xyz[:, atoms[1]] - centroid,
    )
    # Normalize vector
    norm_vec = cross / np.linalg.norm(cross, axis=1)[:, np.newaxis]
    return norm_vec


def dot_pdt(v1, v2):
    """
    Compute the dot product of two stacks of vectors.

    ASsumes v1 and v2 are the same shape.

    Parameters
    ----------
    v1 : ndarray
        The first stack of vectors.
    v2 : ndarray
        The second stack of vectors.

    Returns
    -------
    dot : np.array, shape=v1.shape, dtype=float
        The dot product of the two stacks of vectors.
    """
    dot = np.einsum("ij,ij->i", v1, v2)
    return dot


def compute_angles(vec_1, vec_2):
    """
    Compute the angles between the two supplied vectors.

    Assumes the vectors are the same shape.

    Parameters
    ----------
    vec_1 : ndarray
        The first vector to compute the angles between.
    vec_2 : ndarray
        The second vector to compute the angles between.

    Returns
    -------
    angle : np.array, shape=vec_1.shape, dtype=float
        The angles between the two vectors, in radians.
    """
    angle = np.arccos(dot_pdt(vec_1, vec_2) / (np.linalg.norm(vec_1, axis=1) * np.linalg.norm(vec_2, axis=1)))
    return angle


def cap_angle(angle):
    """
    Bring the angle to the range 0-pi/2.

    For this capping, if the angle is beyond pi/2, we are then functionally finding the
    angle where v2 is the negative of the original.

    Parameters
    ----------
    angle : ndarray
        The vector of angles to cap. Angles assumed to be defined between 0 and pi.

    Returns
    -------
    angle :np.array, shape=angle.shape, dtype=float
        The capped angles. Modifies the input array.
    """
    msk = angle > np.pi / 2
    angle[msk] = np.pi - angle[msk]
    return angle


def calculate_intersection_point(
    plane_normal,
    plane_centroid,
    tilted_normal,
    tilted_centroid,
):
    """
    Calculate the point of intersection between the two potentially T-stacked rings.

    Essentially cargo-culted from ProLIF. This implementation may be improved by using
    the actual planes of the rings instead of the normals to start, since then one
    could just solve for the line of intersection, and then project the centroid
    onto that.

    Parameters
    ----------
    plane_normal : ndarray
        The normal vector of one of the rings.
    plane_centroid : ndarray
        The centroid of one of the rings.
    tilted_normal : ndarray
        The normal vector of the other ring.
    tilted_centroid : ndarray
        The centroid of the other ring.

    Returns
    -------
    points : ndarray
        The point of intersection between the two rings. If the two planes are parallel,
        returns array of NaNs for that frame.
    """
    points = np.zeros(plane_normal.shape)
    intersect_direction = np.cross(plane_normal, tilted_normal, axis=1)
    A = np.stack((plane_normal, tilted_normal, intersect_direction), axis=1)
    dets = np.linalg.det(A)
    singular_matrices = np.isclose(dets, 0)
    A[singular_matrices] = np.random.random((3, 3))
    tilted_offset = dot_pdt(tilted_normal, tilted_centroid)
    plane_offset = dot_pdt(plane_normal, plane_centroid)
    d = np.stack(
        (plane_offset, tilted_offset, np.broadcast_to([0.0], plane_offset.shape)),
        axis=1,
    )
    # NOTE: np.linalg.solve changed in v2.0. This is how to do it in 2.0
    intersect_pt = np.linalg.solve(A, d[..., None])[..., 0]
    vec = plane_centroid - intersect_pt
    intersect_direction = intersect_direction / np.linalg.norm(intersect_direction, axis=1)[:, np.newaxis]
    scalar_proj = dot_pdt(intersect_direction, vec)
    non_singular_matrices = ~singular_matrices
    points[non_singular_matrices] = (
        intersect_pt[non_singular_matrices]
        + intersect_direction[non_singular_matrices] * scalar_proj[non_singular_matrices, np.newaxis]
    )
    points[singular_matrices] = np.array([np.nan, np.nan, np.nan])
    return points


def calculate_face_stack_threshold(
    face_plane_angle_range_rad,
    face_normal_to_centroid_angle_range_rad,
    max_face_to_face_centroid_distance,
    plane_angles,
    res_to_lig_angle,
    lig_to_res_angle,
    centroid_dist,
):
    """
    Calculate the threshold for face-to-face pi-stacking interactions.

    Parameters
    ----------
    face_plane_angle_range_rad : tuple of float
        The range of acceptable angles between the two normal vectors defined by the two
        centroids.
    face_normal_to_centroid_angle_range_rad : tuple of float
        The range of acceptable angles between the normal vector and the vector between
        centroids.
        Checks both the ligand and receptor groups.
    max_face_to_face_centroid_distance : float
        The maximum distance between the centroids of the ligand and receptor groups for
        the interaction.
    plane_angles : ndarray
        The angles between the normals of the two rings. In radians.
    res_to_lig_angle : ndarray
        The angles between the normal of the receptor ring and the vector between the
        centroids. In radians.
    lig_to_res_angle : ndarray
        The angles between the normal of the ligand ring and the vector between the
        centroids. In radians.
    centroid_dist : ndarray
        The distance between the centroids of the two rings. In nm.

    Returns
    -------
    face_msk : ndarray
        The mask of frames that meet the threshold for face-to-face pi-stacking.
    """
    face_msk = (
        (face_plane_angle_range_rad[0] <= plane_angles)
        & (plane_angles <= face_plane_angle_range_rad[1])
        & (
            (
                (face_normal_to_centroid_angle_range_rad[0] <= res_to_lig_angle)
                & (res_to_lig_angle <= face_normal_to_centroid_angle_range_rad[1])
            )
            | (
                (face_normal_to_centroid_angle_range_rad[0] <= lig_to_res_angle)
                & (lig_to_res_angle <= face_normal_to_centroid_angle_range_rad[1])
            )
        )
        & (centroid_dist <= max_face_to_face_centroid_distance)
    )
    return face_msk


def calculate_edge_stack_threshold(
    edge_plane_angle_range_rad,
    edge_normal_to_centroid_angle_range_rad,
    max_edge_to_face_centroid_distance,
    edge_intersection_radius,
    plane_angles,
    res_to_lig_angle,
    lig_to_res_angle,
    centroid_dist,
    min_inter_dist,
):
    """
    Calculate the threshold for edge-to-face pi-stacking interactions.

    Parameters
    ----------
    edge_plane_angle_range_rad : tuple of float
        The range of acceptable angles between the two normal vectors defined by the two
        centroids.
    edge_normal_to_centroid_angle_range_rad : tuple of float
        The range of acceptable angles between the normal vector and the vector between
        centroids. Checks both the ligand and receptor groups.
    max_edge_to_face_centroid_distance : float
        The maximum distance between the centroids of the ligand and receptor groups for
        the interaction.
    edge_intersection_radius : float
        The maximum distance between the point of intersection between both rings and
        the opposite ring's centroid.
    plane_angles : ndarray
        The angles between the normals of the two rings. In radians.
    res_to_lig_angle : ndarray
        The angles between the normal of the receptor ring and the vector between the
        centroids. In radians.
    lig_to_res_angle : ndarray
        The angles between the normal of the ligand ring and the vector between the
        centroids. In radians.
    centroid_dist : ndarray
        The distance between the centroids of the two rings. In nm.
    min_inter_dist : ndarray
        The distance between the intersection point and the centroids of the two rings.
        In nm.

    Returns
    -------
    tstack_msk : np.array, shape=[n_frames], dtype=float
        The mask of frames that meet the threshold for edge-to-face pi-stacking.
    """
    tstack_msk = (
        (edge_plane_angle_range_rad[0] <= plane_angles)
        & (plane_angles <= edge_plane_angle_range_rad[1])
        & (
            (
                (edge_normal_to_centroid_angle_range_rad[0] <= res_to_lig_angle)
                & (res_to_lig_angle <= edge_normal_to_centroid_angle_range_rad[1])
            )
            | (
                (edge_normal_to_centroid_angle_range_rad[0] <= lig_to_res_angle)
                & (lig_to_res_angle <= edge_normal_to_centroid_angle_range_rad[1])
            )
        )
        & (centroid_dist <= max_edge_to_face_centroid_distance)
        & (min_inter_dist <= edge_intersection_radius)
    )
    return tstack_msk


def pi_stacking(
    trajectory,
    ligand_aromatic_groups,
    receptor_aromatic_groups,
    ligand_neighbor_cutoff=None,
    max_face_to_face_centroid_distance=5.5,
    face_plane_angle_range=(0.0, 35.0),
    face_normal_to_centroid_angle_range=(0.0, 33.0),
    max_edge_to_face_centroid_distance=0.65,
    edge_plane_angle_range=(50.0, 90.0),
    edge_normal_to_centroid_angle_range=(0.0, 30.0),
    edge_intersection_radius=0.15,
):
    """
    Calculate the pi-stacking interactions based on supplied atom groups.

    This function calculates the pi-stacking interactions between the ligand and
    receptor groups based on the supplied atom indices. The function uses the
    centroid of the groups to determine the distance between them, and the normal
    vector of the groups to determine the angle between them.

    Two types of pi-stacking interactions are considered:

    1. Face-to-face interactions: The two groups are parallel and close to each other.
    2. Edge-to-face interactions: The two groups are not parallel, but the edge of one
       group is close to the face of the other group.

    Both are returned together in the same list.


    Parameters
    ----------
    trajectory : md.Trajectory
        The trajectory to analyze. The ligand and receptor groups map onto the
          trajectory's topology indices.
    ligand_aromatic_groups : list of int
        The atom indices of the groups to be considered aromatic for the ligand.
    receptor_aromatic_groups : list of int
        The atom indices of the groups to be considered aromatic for the receptor.
    ligand_neighbor_cutoff : float, default=None
        The distance cutoff for considering a receptor group for pi-stacking with a
        ligand group. If None, then all pairwise aromatic groups are considered.
        IMPORTANT: When provided, this filtering is only applied once, using the first
        frame of the trajectory, which may miss interactions that form later due to
        molecular movement. It is recommended that you use this parameter only if:
        (1) your system has minimal conformational changes throughout the trajectory, or
        (2) you're only interested in interactions present in the first frame.
    max_face_to_face_centroid_distance : float
        The maximum distance between the centroids of the ligand and receptor groups for
        the interaction.
    face_plane_angle_range : tuple of float
        The range of acceptable angles between the two normal vectors defined by the two
        centroids. In degrees.
    face_normal_to_centroid_angle_range : tuple of float
        The range of acceptable angles between the normal vector and the vector between
        centroids. Checks both the ligand and receptor groups. In degrees.
    max_edge_to_face_centroid_distance : float
        The maximum distance between the centroids of the ligand and receptor groups for
        the interaction.
    edge_plane_angle_range : tuple of float
        The range of acceptable angles between the two normal vectors defined by the two
        centroids. In degrees.
    edge_normal_to_centroid_angle_range : tuple of float
        The range of acceptable angles between the normal vector and the vector between
        centroids. Checks both the ligand and receptor groups. In degrees.
    edge_intersection_radius : float
        The maximum distance between the point of intersection between both rings and
        the opposite ring's centroid.

    Returns
    -------
    stacking_interactions: list, len=n_frames
        A list of lists of tuples, where each tuple is a pair of aromatic groups that
        are stacking in that frame. The order of the tuple goes
        (ligand_group, protein_group). Includes both face-to-face and edge-to-face
        interactions.
    """
    face_plane_angle_range_rad = tuple(np.deg2rad(face_plane_angle_range))
    face_normal_to_centroid_angle_range_rad = tuple(np.deg2rad(face_normal_to_centroid_angle_range))
    edge_plane_angle_range_rad = tuple(np.deg2rad(edge_plane_angle_range))
    edge_normal_to_centroid_angle_range_rad = tuple(np.deg2rad(edge_normal_to_centroid_angle_range))
    # Ensure order of the groups by converting to list.
    ligand_aromatic_groups = list(ligand_aromatic_groups)
    receptor_aromatic_groups = list(receptor_aromatic_groups)
    # ligand_neighbor_groups = {Ligand group idx: receptor group idxs}
    ligand_neighbor_groups = defaultdict(list)
    if ligand_neighbor_cutoff is not None:
        receptor_nbr_atomatic_groups: set[tuple[int, ...]] = set()
        neighbors = md.compute_neighborlist(
            trajectory,
            ligand_neighbor_cutoff,
        )
        for lig_grp in ligand_aromatic_groups:
            # Find all neighbors for each of the atoms in the ligand group, add to set
            lig_neighbors: set[int] = set()
            for atm_idx in lig_grp:
                lig_neighbors.update(neighbors[atm_idx])
            # For each of the receptor groups, check if any of the atoms are a neighbor
            for rec_grp in receptor_aromatic_groups:
                if any(atm in lig_neighbors for atm in rec_grp):
                    ligand_neighbor_groups[lig_grp].append(rec_grp)
                    receptor_nbr_atomatic_groups.add(rec_grp)
    else:
        receptor_nbr_atomatic_groups = set(receptor_aromatic_groups)
        for rec_grp in receptor_aromatic_groups:
            for lig_grp in ligand_aromatic_groups:
                ligand_neighbor_groups[lig_grp].append(rec_grp)
    stacking_interactions = [[] for _ in range(len(trajectory))]
    # If no receptor aromatic neighbors within any of the ligand groups, return
    if len(ligand_neighbor_groups) == 0:
        return stacking_interactions
    # Calculate the centroid and ring normals of each group
    receptor_grp_centroids = {}
    receptor_grp_normals = {}
    for rec_grp in receptor_nbr_atomatic_groups:
        rec_centroid = compute_centroid(trajectory, rec_grp)
        receptor_grp_centroids[rec_grp] = rec_centroid
        receptor_grp_normals[rec_grp] = compute_ring_normal(trajectory, rec_grp, rec_centroid)
    # For each ligand -- receptor pairs, calculate the centroid distance
    for lig_grp, receptor_grps in ligand_neighbor_groups.items():
        lig_centroid = compute_centroid(
            trajectory,
            lig_grp,
        )
        lig_normal = compute_ring_normal(
            trajectory,
            lig_grp,
            lig_centroid,
        )
        # Calculate the plane angles
        for grp in receptor_grps:
            receptor_centroid = receptor_grp_centroids[grp]
            centroid_dist = compute_distances_core(
                np.hstack(
                    (lig_centroid[:, np.newaxis], receptor_centroid[:, np.newaxis]),
                ),
                [[0, 1]],
                trajectory.unitcell_vectors,
            ).squeeze()
            if np.all(
                centroid_dist
                > max(
                    max_face_to_face_centroid_distance,
                    max_edge_to_face_centroid_distance,
                ),
            ):
                # centroid distance too far for all frames, skip
                continue
            # Calculate the angle between the normals
            # Going to assume PBC doesn't apply here since we're just using unit vectors
            res_normal = receptor_grp_normals[grp]
            plane_angles = cap_angle(compute_angles(lig_normal, res_normal))
            # Calcualte the normal vec - centroid vec angles
            res_to_lig_centroid = lig_centroid - receptor_centroid
            res_to_lig_angle = cap_angle(compute_angles(res_normal, res_to_lig_centroid))
            lig_to_res_centroid = receptor_centroid - lig_centroid
            lig_to_res_angle = cap_angle(compute_angles(lig_normal, lig_to_res_centroid))
            lig_res_frames_msk = calculate_face_stack_threshold(
                face_plane_angle_range_rad,
                face_normal_to_centroid_angle_range_rad,
                max_face_to_face_centroid_distance,
                plane_angles,
                res_to_lig_angle,
                lig_to_res_angle,
                centroid_dist,
            )
            for frame in np.where(lig_res_frames_msk)[0]:
                stacking_interactions[frame].append((lig_grp, grp))

            # T-stacks
            intersections = calculate_intersection_point(
                lig_normal,
                lig_centroid,
                res_normal,
                receptor_centroid,
            )
            lig_inter_dist = np.linalg.norm(lig_centroid - intersections, axis=1)
            rec_inter_dist = np.linalg.norm(receptor_centroid - intersections, axis=1)
            min_inter_dist = np.fmin(
                lig_inter_dist,
                rec_inter_dist,
            )
            tstack_msk = calculate_edge_stack_threshold(
                edge_plane_angle_range_rad,
                edge_normal_to_centroid_angle_range_rad,
                max_edge_to_face_centroid_distance,
                edge_intersection_radius,
                plane_angles,
                res_to_lig_angle,
                lig_to_res_angle,
                centroid_dist,
                min_inter_dist,
            )
            for frame in np.where(tstack_msk)[0]:
                stacking_interactions[frame].append((lig_grp, grp))
    return stacking_interactions
